import atexit
import contextlib
import logging
import os
import pathlib
import pprint
import re
import shutil
import subprocess
import textwrap
import time
from typing import TYPE_CHECKING

import attr
import distro
import packaging.version
import psutil
import pytest
import requests
import saltfactories.cli
from pytestshellutils.shell import DaemonImpl, Subprocess
from pytestshellutils.utils.processes import (
    ProcessResult,
    _get_cmdline,
    terminate_process,
    terminate_process_list,
)
from pytestskipmarkers.utils import platform
from saltfactories.bases import SystemdSaltDaemonImpl
from saltfactories.cli import call, key
from saltfactories.daemons import api, master, minion
from saltfactories.utils import cli_scripts

import salt.utils.files
from tests.conftest import CODE_DIR
from tests.support.pytest.helpers import TestAccount, download_file

ARTIFACTS_DIR = CODE_DIR / "artifacts" / "pkg"

log = logging.getLogger(__name__)


@attr.s(kw_only=True, slots=True)
class SaltPkgInstall:
    pkg_system_service: bool = attr.ib(default=False)
    proc: Subprocess = attr.ib(init=False, repr=False)

    # Paths
    root: pathlib.Path = attr.ib(default=None)
    run_root: pathlib.Path = attr.ib(default=None)
    ssm_bin: pathlib.Path = attr.ib(default=None)
    bin_dir: pathlib.Path = attr.ib(default=None)
    install_dir: pathlib.Path = attr.ib(init=False)
    binary_paths: dict[str, list[pathlib.Path]] = attr.ib(init=False)
    config_path: str = attr.ib(init=False)
    conf_dir: pathlib.Path = attr.ib()

    # Test selection flags
    upgrade: bool = attr.ib(default=False)
    downgrade: bool = attr.ib(default=False)
    classic: bool = attr.ib(default=False)

    # Installing flags
    no_install: bool = attr.ib(default=False)
    no_uninstall: bool = attr.ib(default=False)

    # Distribution/system information
    distro_id: str = attr.ib(init=False)
    distro_codename: str = attr.ib(init=False)
    distro_name: str = attr.ib(init=False)
    distro_version: str = attr.ib(init=False)

    # Version information
    prev_version: str = attr.ib()
    use_prev_version: str = attr.ib()
    artifact_version: str = attr.ib(init=False)
    version: str = attr.ib(init=False)

    # Package (and management) metadata
    pkg_mngr: str = attr.ib(init=False)
    rm_pkg: str = attr.ib(init=False)
    dbg_pkg: str = attr.ib(init=False)
    salt_pkgs: list[str] = attr.ib(init=False)
    pkgs: list[str] = attr.ib(factory=list)
    file_ext: bool = attr.ib(default=None)
    relenv: bool = attr.ib(default=True)

    @proc.default
    def _default_proc(self):
        return Subprocess(timeout=240)

    @distro_id.default
    def _default_distro_id(self):
        return distro.id().lower()

    @distro_codename.default
    def _default_distro_codename(self):
        return distro.codename().lower()

    @distro_name.default
    def _default_distro_name(self):
        name = distro.name()
        if name:
            if "vmware" in name.lower():
                return name.split()[1].lower()
            return name.split()[0].lower()

    @distro_version.default
    def _default_distro_version(self):
        if self.distro_name in ("photon", "rocky"):
            return distro.version().split(".")[0]
        return distro.version().lower()

    @pkg_mngr.default
    def _default_pkg_mngr(self):
        if self.distro_id in (
            "almalinux",
            "rocky",
            "centos",
            "redhat",
            "amzn",
            "fedora",
            "photon",
        ):
            return "yum"
        elif self.distro_id in ("ubuntu", "debian"):
            ret = self.proc.run("apt-get", "update")
            self._check_retcode(ret)
            return "apt-get"

    @rm_pkg.default
    def _default_rm_pkg(self):
        if self.distro_id in (
            "almalinux",
            "rocky",
            "centos",
            "redhat",
            "amzn",
            "fedora",
            "photon",
        ):
            return "remove"
        elif self.distro_id in ("ubuntu", "debian"):
            return "purge"

    @dbg_pkg.default
    def _default_dbg_pkg(self):
        dbg_pkg = None
        if self.distro_id in (
            "almalinux",
            "rocky",
            "centos",
            "redhat",
            "amzn",
            "fedora",
            "photon",
        ):
            dbg_pkg = "salt-debuginfo"
        elif self.distro_id in ("ubuntu", "debian"):
            dbg_pkg = "salt-dbg"
        return dbg_pkg

    @salt_pkgs.default
    def _default_salt_pkgs(self):
        salt_pkgs = [
            "salt-api",
            "salt-syndic",
            "salt-ssh",
            "salt-master",
            "salt-cloud",
            "salt-minion",
        ]
        if self.distro_id in (
            "almalinux",
            "rocky",
            "centos",
            "redhat",
            "amzn",
            "fedora",
            "photon",
        ):
            salt_pkgs.append("salt")
        elif self.distro_id in ("ubuntu", "debian"):
            salt_pkgs.append("salt-common")
        if packaging.version.parse(self.version) >= packaging.version.parse("3006.3"):
            if self.dbg_pkg:
                salt_pkgs.append(self.dbg_pkg)
        return salt_pkgs

    @install_dir.default
    def _default_install_dir(self):
        if platform.is_windows():
            install_dir = pathlib.Path(
                os.getenv("ProgramFiles"), "Salt Project", "Salt"
            ).resolve()
        elif platform.is_darwin():
            install_dir = pathlib.Path("/opt", "salt")
        else:
            install_dir = pathlib.Path("/opt", "saltstack", "salt")
        return install_dir

    @config_path.default
    def _default_config_path(self):
        """
        Default location for salt configurations
        """
        if platform.is_windows():
            config_path = pathlib.Path("C:\\salt", "etc", "salt")
        else:
            config_path = pathlib.Path("/etc", "salt")
        return config_path

    @version.default
    def _default_version(self):
        """
        The version to be installed at the start
        """
        if not self.upgrade and not self.use_prev_version:
            version = self.artifact_version
        else:
            version = self.prev_version
            parsed = packaging.version.parse(version)
            version = f"{parsed.major}.{parsed.minor}"
        # ensure services stopped on Debian/Ubuntu (minic install for RedHat - non-starting)
        if self.distro_id in ("ubuntu", "debian"):
            self.stop_services()
        return version

    @artifact_version.default
    def _default_artifact_version(self):
        """
        The version of the local salt artifacts being tested, based on regex matching
        """
        version = ""
        artifacts = list(ARTIFACTS_DIR.glob("**/*.*"))
        for artifact in artifacts:
            version = re.search(
                r"([0-9].*)(\-[0-9].fc|\-[0-9].el|\+ds|\_all|\_any|\_amd64|\_arm64|\-[0-9].am|(\-[0-9]-[a-z]*-[a-z]*[0-9_]*.|\-[0-9]*.*)(exe|msi|pkg|rpm|deb))",
                artifact.name,
            )
            if version:
                version = version.groups()[0].replace("_", "-").replace("~", "")
                version = version.split("-")[0]
                break
        if not version:
            pytest.fail(
                f"Failed to find package artifacts in '{ARTIFACTS_DIR}'. "
                f"Directory Contents:\n{pprint.pformat(artifacts)}"
            )
        return version

    def update_process_path(self):
        # The installer updates the path for the system, but that doesn't
        # make it to this python session, so we need to update that
        os.environ["PATH"] = ";".join([str(self.install_dir), os.getenv("path")])

    def __attrs_post_init__(self):
        self.relenv = packaging.version.parse(self.version) >= packaging.version.parse(
            "3006.0"
        )

        file_ext_re = "rpm|deb"
        if platform.is_darwin():
            file_ext_re = "pkg"
        if platform.is_windows():
            file_ext_re = "exe|msi"

        for f_path in ARTIFACTS_DIR.glob("**/*.*"):
            f_path = str(f_path)
            if re.search(f"salt-(.*).({file_ext_re})$", f_path, re.IGNORECASE):
                self.file_ext = os.path.splitext(f_path)[1].strip(".")
                self.pkgs.append(f_path)
                if platform.is_windows():
                    self.root = pathlib.Path(os.getenv("LocalAppData")).resolve()
                    if self.file_ext in ["exe", "msi"]:
                        self.root = self.install_dir.parent
                        self.bin_dir = self.install_dir
                        self.ssm_bin = self.install_dir / "ssm.exe"
                        self.run_root = self.bin_dir / "bin" / "salt.exe"
                        if not self.relenv and not self.classic:
                            self.ssm_bin = self.bin_dir / "bin" / "ssm.exe"
                    else:
                        log.error("Unexpected file extension: %s", self.file_ext)
                    if self.use_prev_version:
                        self.bin_dir = self.install_dir / "bin"
                        self.run_root = self.bin_dir / "salt.exe"
                        self.ssm_bin = self.bin_dir / "ssm.exe"
                        if self.file_ext == "msi" or self.relenv:
                            self.ssm_bin = self.install_dir / "ssm.exe"
                        if (
                            self.install_dir / "salt-minion.exe"
                        ).exists() and not self.relenv:
                            log.debug(
                                "Removing %s", self.install_dir / "salt-minion.exe"
                            )
                            (self.install_dir / "salt-minion.exe").unlink()

                elif platform.is_darwin():
                    self.root = pathlib.Path("/opt")
                    if self.file_ext == "pkg":
                        self.bin_dir = self.root / "salt" / "bin"
                        self.run_root = self.bin_dir / "run"
                    else:
                        log.error("Unexpected file extension: %s", self.file_ext)
                log.debug("root: %s", self.root)
                log.debug("bin_dir: %s", self.bin_dir)
                log.debug("ssm_bin: %s", self.ssm_bin)
                log.debug("run_root: %s", self.run_root)

        if not self.pkgs:
            pytest.fail("Could not find Salt Artifacts")

        python_bin = self.install_dir / "bin" / "python3"
        if platform.is_windows():
            python_bin = self.install_dir / "Scripts" / "python.exe"
            if self.relenv:
                self.binary_paths = {
                    "call": ["salt-call.exe"],
                    "cp": ["salt-cp.exe"],
                    "minion": ["salt-minion.exe"],
                    "pip": ["salt-pip.exe"],
                    "python": [python_bin],
                }
            elif self.classic:
                self.binary_paths = {
                    "call": [self.install_dir / "salt-call.bat"],
                    "cp": [self.install_dir / "salt-cp.bat"],
                    "minion": [self.install_dir / "salt-minion.bat"],
                    "python": [self.bin_dir / "python.exe"],
                }
                self.binary_paths["pip"] = self.binary_paths["python"] + ["-m", "pip"]
            else:
                self.binary_paths = {
                    "call": [str(self.run_root), "call"],
                    "cp": [str(self.run_root), "cp"],
                    "minion": [str(self.run_root), "minion"],
                    "pip": [str(self.run_root), "pip"],
                    "python": [str(self.run_root), "shell"],
                }

        else:
            if os.path.exists(self.install_dir / "bin" / "salt"):
                install_dir = self.install_dir / "bin"
            else:
                install_dir = self.install_dir
            if self.relenv:
                self.binary_paths = {
                    "salt": [install_dir / "salt"],
                    "api": [install_dir / "salt-api"],
                    "call": [install_dir / "salt-call"],
                    "cloud": [install_dir / "salt-cloud"],
                    "cp": [install_dir / "salt-cp"],
                    "key": [install_dir / "salt-key"],
                    "master": [install_dir / "salt-master"],
                    "minion": [install_dir / "salt-minion"],
                    "proxy": [install_dir / "salt-proxy"],
                    "run": [install_dir / "salt-run"],
                    "ssh": [install_dir / "salt-ssh"],
                    "syndic": [install_dir / "salt-syndic"],
                    "spm": [install_dir / "spm"],
                    "pip": [install_dir / "salt-pip"],
                    "python": [python_bin],
                }
            else:
                self.binary_paths = {
                    "salt": [shutil.which("salt")],
                    "api": [shutil.which("salt-api")],
                    "call": [shutil.which("salt-call")],
                    "cloud": [shutil.which("salt-cloud")],
                    "cp": [shutil.which("salt-cp")],
                    "key": [shutil.which("salt-key")],
                    "master": [shutil.which("salt-master")],
                    "minion": [shutil.which("salt-minion")],
                    "proxy": [shutil.which("salt-proxy")],
                    "run": [shutil.which("salt-run")],
                    "ssh": [shutil.which("salt-ssh")],
                    "syndic": [shutil.which("salt-syndic")],
                    "spm": [shutil.which("spm")],
                    "python": [str(pathlib.Path("/usr/bin/python3"))],
                }
                if self.classic:
                    if platform.is_darwin():
                        # `which` is not catching the right paths on downgrades, explicitly defining them here
                        self.binary_paths = {
                            "salt": [self.bin_dir / "salt"],
                            "api": [self.bin_dir / "salt-api"],
                            "call": [self.bin_dir / "salt-call"],
                            "cloud": [self.bin_dir / "salt-cloud"],
                            "cp": [self.bin_dir / "salt-cp"],
                            "key": [self.bin_dir / "salt-key"],
                            "master": [self.bin_dir / "salt-master"],
                            "minion": [self.bin_dir / "salt-minion"],
                            "proxy": [self.bin_dir / "salt-proxy"],
                            "run": [self.bin_dir / "salt-run"],
                            "ssh": [self.bin_dir / "salt-ssh"],
                            "syndic": [self.bin_dir / "salt-syndic"],
                            "spm": [self.bin_dir / "spm"],
                            "python": [str(self.bin_dir / "python3")],
                            "pip": [str(self.bin_dir / "pip3")],
                        }
                    else:
                        self.binary_paths["pip"] = [str(pathlib.Path("/usr/bin/pip3"))]
                        self.proc.run(*self.binary_paths["pip"], "install", "-U", "pip")
                        self.proc.run(
                            *self.binary_paths["pip"], "install", "-U", "pyopenssl"
                        )
                else:
                    self.binary_paths["python"] = [shutil.which("salt"), "shell"]
                    if platform.is_darwin():
                        self.binary_paths["pip"] = [self.run_root, "pip"]
                        self.binary_paths["spm"] = [shutil.which("salt-spm")]
                    else:
                        self.binary_paths["pip"] = [shutil.which("salt-pip")]
        log.debug("python_bin: %s", python_bin)
        log.debug("binary_paths: %s", self.binary_paths)
        log.debug("install_dir: %s", self.install_dir)

    @staticmethod
    def salt_factories_root_dir(system_service: bool = False) -> pathlib.Path:
        if system_service is False:
            return None
        if platform.is_windows():
            return pathlib.Path("C:\\salt")
        if platform.is_darwin():
            return pathlib.Path("/opt/salt")
        return pathlib.Path("/")

    def _check_retcode(self, ret):
        """
        Helper function to check subprocess.run returncode equals 0
        If not raise AssertionError
        """
        if ret.returncode != 0:
            log.error(ret)
        assert ret.returncode == 0
        return True

    def _install_pkgs(self, upgrade=False, downgrade=False):
        if downgrade:
            self.install_previous(downgrade=downgrade)
            return True
        pkg = str(pathlib.Path(self.pkgs[0]).resolve())
        if platform.is_windows():
            if upgrade:
                self.root = self.install_dir.parent
                self.bin_dir = self.install_dir
                self.ssm_bin = self.install_dir / "ssm.exe"
            if pkg.endswith("exe"):
                # Install the package
                log.info("Installing: %s", str(pkg))
                ret = self.proc.run(str(pkg), "/start-minion=0", "/S")
                self._check_retcode(ret)
            elif pkg.endswith("msi"):
                # Install the package
                log.info("Installing: %s", str(pkg))
                # self.proc.run always makes the command a list even when shell
                # is true, meaning shell being true will never work correctly.
                ret = subprocess.run(
                    f'msiexec.exe /qn /i {pkg} /norestart START_MINION=""',
                    shell=True,  # nosec
                    check=False,
                )
                assert ret.returncode in [0, 3010]
            else:
                log.error("Invalid package: %s", pkg)
                return False

            # Remove the service installed by the installer
            log.debug("Removing installed salt-minion service")
            self.proc.run(str(self.ssm_bin), "remove", "salt-minion", "confirm")

            # Add installation to the path
            self.update_process_path()

            # Install the service using our config
            if self.pkg_system_service:
                self._install_ssm_service()

        elif platform.is_darwin():
            daemons_dir = pathlib.Path("/Library", "LaunchDaemons")
            service_name = "com.saltstack.salt.minion"
            plist_file = daemons_dir / f"{service_name}.plist"
            log.debug("Installing: %s", str(pkg))
            ret = self.proc.run("installer", "-pkg", str(pkg), "-target", "/")
            self._check_retcode(ret)

            # Stop the service installed by the installer
            self.proc.run("launchctl", "disable", f"system/{service_name}")
            self.proc.run("launchctl", "bootout", "system", str(plist_file))

        elif upgrade:
            env = os.environ.copy()
            extra_args = []
            if self.distro_id in ("ubuntu", "debian"):

                pref_file = pathlib.Path(
                    "/etc", "apt", "preferences.d", "salt-pin-1001"
                )
                pref_file.parent.mkdir(exist_ok=True)
                pin = f"{self.artifact_version.rsplit('.', 1)[0]}.*"
                with salt.utils.files.fopen(pref_file, "w") as fp:
                    fp.write(
                        f"Package: salt-*\n"
                        f"Pin: version {pin}\n"
                        f"Pin-Priority: 1001"
                    )
                log.error("Pin to %s", pin)

                env["DEBIAN_FRONTEND"] = "noninteractive"
                extra_args = [
                    "-o",
                    "DPkg::Options::=--force-confdef",
                    "-o",
                    "DPkg::Options::=--force-confold",
                ]
            log.info("Installing packages:\n%s", pprint.pformat(self.pkgs))
            args = extra_args + self.pkgs
            upgrade_cmd = "upgrade"
            if self.distro_id == "photon":
                # tdnf does not detect nightly build versions to be higher version
                # than release versions
                upgrade_cmd = "install"
            ret = self.proc.run(
                self.pkg_mngr,
                upgrade_cmd,
                "-y",
                *args,
                env=env,
            )
        else:
            if self.distro_id == "photon":
                ret = self.proc.run(
                    "rpm",
                    "--import",
                    "https://packages.broadcom.com/artifactory/api/security/keypair/SaltProjectKey/public",
                )
                self._check_retcode(ret)
            log.info("Installing packages:\n%s", pprint.pformat(self.pkgs))
            ret = self.proc.run(self.pkg_mngr, "install", "-y", *self.pkgs)

        if not platform.is_darwin() and not platform.is_windows():
            # Make sure we don't have any trailing references to old package file locations
            assert ret.returncode == 0
            assert "/saltstack/salt/run" not in ret.stdout
            log.info(ret)
            self._check_retcode(ret)

    def _install_ssm_service(self, service="minion"):
        """
        This function installs the service on Windows using SSM but does not
        start it.

        Args:

            service (str):
                The name of the service. Default is ``minion``
        """
        service_name = f"salt-{service}"
        binary = self.install_dir / f"{service_name}.exe"
        ret = self.proc.run(
            str(self.ssm_bin),
            "install",
            service_name,
            binary,
            "-c",
            f'"{str(self.conf_dir)}"',
        )
        self._check_retcode(ret)
        ret = self.proc.run(
            str(self.ssm_bin),
            "set",
            service_name,
            "Description",
            "Salt Minion for testing",
        )
        self._check_retcode(ret)
        # This doesn't start the service. It will start automatically on reboot
        # It is set here to make it the same as what the installer does
        ret = self.proc.run(
            str(self.ssm_bin), "set", service_name, "Start", "SERVICE_AUTO_START"
        )
        self._check_retcode(ret)
        ret = self.proc.run(
            str(self.ssm_bin), "set", service_name, "AppStopMethodConsole", "24000"
        )
        self._check_retcode(ret)
        ret = self.proc.run(
            str(self.ssm_bin), "set", service_name, "AppStopMethodWindow", "2000"
        )
        self._check_retcode(ret)
        ret = self.proc.run(
            str(self.ssm_bin), "set", service_name, "AppRestartDelay", "60000"
        )
        self._check_retcode(ret)

    def package_python_version(self):
        return self.proc.run(
            str(self.binary_paths["python"][0]),
            "-c",
            "import sys; print('{}.{}'.format(*sys.version_info))",
        ).stdout.strip()

    def install(self, upgrade=False, downgrade=False):
        self._install_pkgs(upgrade=upgrade, downgrade=downgrade)
        if self.distro_id in ("ubuntu", "debian"):
            self.stop_services()

    def stop_services(self):
        """
        Debian/Ubuntu distros automatically start the services on install
        We want to ensure our tests start with the config settings we have set.
        This will also verify the expected services are up and running.
        """
        retval = True
        for service in ["salt-syndic", "salt-master", "salt-minion"]:
            check_run = self.proc.run("systemctl", "status", service)
            if check_run.returncode != 0:
                # The system was not started automatically and
                # we are expecting it to be on install on Debian/Ubuntu systems
                log.debug("The service %s was not started on install.", service)
                retval = False
            else:
                stop_service = self.proc.run("systemctl", "stop", service)
                self._check_retcode(stop_service)
        return retval

    def restart_services(self):
        """
        Debian/Ubuntu distros automatically start the services
        We want to ensure our tests start with the config settings we have set,
        for example: after install the services are stopped (similar to RedHat not starting services on install)
        This will also verify the expected services are up and running.
        """
        for service in ["salt-minion", "salt-master", "salt-syndic"]:
            check_run = self.proc.run("systemctl", "status", service)
            log.debug(
                "The restart_services status, before restart, for service %s is %s.",
                service,
                check_run,
            )
            restart_service = self.proc.run("systemctl", "restart", service)
            self._check_retcode(restart_service)

    def install_previous(self, downgrade=False):
        """
        Install previous version. This is used for upgrade tests.
        """
        major_ver = packaging.version.parse(self.prev_version).major
        relenv = packaging.version.parse(self.prev_version) >= packaging.version.parse(
            "3006.0"
        )
        distro_name = self.distro_name
        if distro_name in ("almalinux", "rocky", "centos", "fedora"):
            distro_name = "redhat"
        root_url = "https://packages.broadcom.com/artifactory"

        if self.distro_name in [
            "almalinux",
            "rocky",
            "redhat",
            "centos",
            "amazon",
            "fedora",
            "vmware",
            "photon",
        ]:
            # Removing EPEL repo files
            for fp in pathlib.Path("/etc", "yum.repos.d").glob("epel*"):
                fp.unlink()

            if platform.is_aarch64():
                arch = "arm64"
                # Starting with 3006.5, we prioritize the aarch64 repo paths for rpm-based distros
                if packaging.version.parse(
                    self.prev_version
                ) >= packaging.version.parse("3006.5"):
                    arch = "aarch64"
            else:
                arch = "x86_64"
            ret = self.proc.run(
                "rpm",
                "--import",
                "https://packages.broadcom.com/artifactory/api/security/keypair/SaltProjectKey/public",
            )
            self._check_retcode(ret)
            download_file(
                "https://github.com/saltstack/salt-install-guide/releases/latest/download/salt.repo",
                f"/etc/yum.repos.d/salt-{distro_name}.repo",
            )

            cmd_action = "downgrade" if downgrade else "install"
            pkgs_to_install = self.salt_pkgs.copy()

            if self.distro_name == "photon":
                orig_pkgs = pkgs_to_install[:]
                pkgs_to_install = []
                for _ in orig_pkgs:
                    pkgs_to_install.append(f"{_}-{self.prev_version}")
                ret = self.proc.run(self.pkg_mngr, "clean", "all")
                self._check_retcode(ret)
            else:
                if "3007" in self.prev_version:
                    ret = self.proc.run(
                        self.pkg_mngr,
                        "config-manager",
                        "--enable",
                        "salt-repo-3007-sts",
                    )
                    self._check_retcode(ret)
                else:
                    ret = self.proc.run(
                        self.pkg_mngr,
                        "config-manager",
                        "--disable",
                        "salt-repo-3007-sts",
                    )
                    self._check_retcode(ret)
                ret = self.proc.run(self.pkg_mngr, "clean", "expire-cache")
                self._check_retcode(ret)

            if self.distro_version == "8" and self.classic:
                # centosstream 8 doesn't downgrade properly using the downgrade command for some reason
                # So we explicitly install the correct version here
                list_ret = self.proc.run(
                    self.pkg_mngr, "list", "--available", "salt"
                ).stdout.split("\n")
                list_ret = [_.strip() for _ in list_ret]
                idx = list_ret.index("Available Packages")
                old_ver = list_ret[idx + 1].split()[1]
                pkgs_to_install = [f"{pkg}-{old_ver}" for pkg in pkgs_to_install]
                if self.dbg_pkg:
                    # self.dbg_pkg does not exist on classic packages
                    dbg_exists = [x for x in pkgs_to_install if self.dbg_pkg in x]
                    if dbg_exists:
                        pkgs_to_install.remove(dbg_exists[0])
                cmd_action = "install"
            # pkgs = [f"{_}=={self.prev_version}" for _ in pkgs_to_install]
            ret = self.proc.run(
                self.pkg_mngr,
                cmd_action,
                *pkgs_to_install,
                "-y",
            )
            log.error("**WTF %r", ret)
            self._check_retcode(ret)

        elif distro_name in ["debian", "ubuntu"]:
            ret = self.proc.run(self.pkg_mngr, "install", "curl", "-y")
            self._check_retcode(ret)
            ret = self.proc.run(self.pkg_mngr, "install", "apt-transport-https", "-y")
            self._check_retcode(ret)
            ## only classic 3005 has arm64 support
            if relenv and platform.is_aarch64():
                arch = "arm64"
            elif platform.is_aarch64() and self.classic:
                arch = "arm64"
            else:
                arch = "amd64"

            pathlib.Path("/etc/apt/keyrings").mkdir(parents=True, exist_ok=True)
            gpg_full_path = "/etc/apt/keyrings/salt-archive-keyring.pgp"

            # download the gpg pub key
            download_file(
                f"{root_url}/api/security/keypair/SaltProjectKey/public",
                f"{gpg_full_path}",
            )
            with salt.utils.files.fopen(
                pathlib.Path("/etc", "apt", "sources.list.d", "salt.list"), "w"
            ) as fp:
                fp.write(
                    f"deb [signed-by={gpg_full_path} arch={arch}] "
                    f"{root_url}/saltproject-deb/ stable main"
                )
            self._check_retcode(ret)
            pref_file = pathlib.Path("/etc", "apt", "preferences.d", "salt-pin-1001")
            pref_file.parent.mkdir(exist_ok=True)
            pin = f"{self.prev_version.rsplit('.', 1)[0]}.*"
            if downgrade:
                pin = self.prev_version
            with salt.utils.files.fopen(pref_file, "w") as fp:
                fp.write(
                    f"Package: salt-*\n" f"Pin: version {pin}\n" f"Pin-Priority: 1001"
                )

            cmd = [self.pkg_mngr, "install", *self.salt_pkgs, "-y"]

            # if downgrade:
            #    pref_file = pathlib.Path("/etc", "apt", "preferences.d", "salt-pin-1001")
            #    pref_file.parent.mkdir(exist_ok=True)
            #    # TODO: There's probably something I should put in here to say what version
            #    # TODO: But maybe that's done elsewhere, hopefully in self.salt_pkgs
            #    pref_file.write_text(
            #        textwrap.dedent(
            #            f"""\
            #    Package: salt*
            #    Pin: origin "{root_url}/saltproject-deb"
            #    Pin-Priority: 1001
            #    """
            #        ),
            #        encoding="utf-8",
            #    )
            cmd.append("--allow-downgrades")
            env = os.environ.copy()
            env["DEBIAN_FRONTEND"] = "noninteractive"
            extra_args = [
                "-o",
                "DPkg::Options::=--force-confdef",
                "-o",
                "DPkg::Options::=--force-confold",
            ]
            self.proc.run(self.pkg_mngr, "update", *extra_args, env=env)

            cmd.extend(extra_args)
            log.error("Run cmd %s", cmd)
            ret = self.proc.run(*cmd, env=env)
            log.error("cmd return %r", ret)
            # Pre-relenv packages down get downgraded to cleanly programmatically
            # They work manually, and the install tests after downgrades will catch problems with the install
            self._check_retcode(ret)
            # Let's not check the returncode if this is the case
            # if not (
            #    downgrade
            #    and packaging.version.parse(self.prev_version)
            #    < packaging.version.parse("3006.0")
            # ):
            #    self._check_retcode(ret)
            if downgrade and not self.no_uninstall:
                pref_file.unlink()
            self.stop_services()
        elif platform.is_windows():
            self.bin_dir = self.install_dir / "bin"
            self.run_root = self.bin_dir / "salt.exe"
            self.ssm_bin = self.install_dir / "ssm.exe"
            pkg = str(pathlib.Path(self.pkgs[0]).resolve())

            if self.file_ext == "exe":
                win_pkg = (
                    f"Salt-Minion-{self.prev_version}-Py3-AMD64-Setup.{self.file_ext}"
                )
            elif self.file_ext == "msi":
                win_pkg = f"Salt-Minion-{self.prev_version}-Py3-AMD64.{self.file_ext}"
            else:
                log.debug("Unknown windows file extension: %s", self.file_ext)

            win_pkg_url = (
                f"{root_url}/saltproject-generic/windows/{self.prev_version}/{win_pkg}"
            )
            pkg_path = pathlib.Path(r"C:\TEMP", win_pkg)
            pkg_path.parent.mkdir(exist_ok=True)
            download_file(win_pkg_url, pkg_path)

            if self.file_ext == "msi":

                if downgrade:
                    # MSI can not be downgraded, we must remove the newer version
                    # before installing the old one.
                    ret = subprocess.run(
                        f"msiexec.exe /qn /x {pkg} /norestart",
                        shell=True,  # nosec
                        check=False,
                    )
                    assert ret.returncode == 0

                # self.proc.run always makes the command a list even when shell
                # is true, meaning shell being true will never work correctly.
                ret = subprocess.run(
                    f'msiexec.exe /qn /i {pkg_path} /norestart START_MINION=""',
                    shell=True,  # nosec
                    check=False,
                )
                assert ret.returncode in [0, 3010]
            else:
                # ret = self.proc.run("start", "/wait", f"\"{pkg_path} /start-minion=0 /S\"")
                batch_file = pkg_path.parent / "install_nsis.cmd"
                batch_content = f"start /wait {str(pkg_path)} /start-minion=0 /S"
                with salt.utils.files.fopen(batch_file, "w") as fp:
                    fp.write(batch_content)
                # Now run the batch file
                ret = self.proc.run("cmd.exe", "/c", str(batch_file))
                self._check_retcode(ret)

            log.debug("Removing installed salt-minion service")
            ret = self.proc.run(str(self.ssm_bin), "remove", "salt-minion", "confirm")
            self._check_retcode(ret)

            # Add installation to the path
            self.update_process_path()

            if self.pkg_system_service:
                self._install_ssm_service()

        elif platform.is_darwin():
            if relenv and platform.is_aarch64():
                arch = "arm64"
            elif platform.is_aarch64() and self.classic:
                arch = "arm64"
            else:
                arch = "x86_64"

            mac_pkg = f"salt-{self.prev_version}-py3-{arch}.pkg"
            mac_pkg_url = (
                f"{root_url}/saltproject-generic/macos/{self.prev_version}/{mac_pkg}"
            )

            mac_pkg_path = f"/tmp/{mac_pkg}"
            if not os.path.exists(mac_pkg_path):
                download_file(
                    f"{mac_pkg_url}",
                    f"/tmp/{mac_pkg}",
                )

            ret = self.proc.run("installer", "-pkg", mac_pkg_path, "-target", "/")
            self._check_retcode(ret)

    def uninstall(self):
        pkg = self.pkgs[0]
        if platform.is_windows():
            log.info("Uninstalling %s", pkg)
            if pkg.endswith("exe"):
                uninst = self.install_dir / "uninst.exe"
                ret = self.proc.run(uninst, "/S")
                self._check_retcode(ret)
            elif pkg.endswith("msi"):
                ret = self.proc.run("msiexec.exe", "/qn", "/x", pkg)
                self._check_retcode(ret)

        elif platform.is_darwin():
            # From here: https://stackoverflow.com/a/46118276/4581998
            daemons_dir = pathlib.Path("/Library", "LaunchDaemons")
            for service in ("minion", "master", "api", "syndic"):
                service_name = f"com.saltstack.salt.{service}"
                plist_file = daemons_dir / f"{service_name}.plist"
                # Stop the services
                self.proc.run("launchctl", "disable", f"system/{service_name}")
                self.proc.run("launchctl", "bootout", "system", str(plist_file))

            # Remove Symlink to salt-config
            if os.path.exists("/usr/local/sbin/salt-config"):
                os.unlink("/usr/local/sbin/salt-config")

            # Remove supporting files
            self.proc.run(
                "pkgutil",
                "--only-files",
                "--files",
                "com.saltstack.salt",
                "|",
                "grep",
                "-v",
                "opt",
                "|",
                "tr",
                "'\n'",
                "' '",
                "|",
                "xargs",
                "-0",
                "rm",
                "-f",
            )

            # Remove directories
            if os.path.exists("/etc/salt"):
                shutil.rmtree("/etc/salt")

            # Remove path
            if os.path.exists("/etc/paths.d/salt"):
                os.remove("/etc/paths.d/salt")

            # Remove receipt
            self.proc.run("pkgutil", "--forget", "com.saltstack.salt")

            log.debug("Deleting the onedir directory: %s", self.root / "salt")
            shutil.rmtree(str(self.root / "salt"))
        else:
            log.debug("Un-Installing packages:\n%s", pprint.pformat(self.salt_pkgs))
            ret = self.proc.run(self.pkg_mngr, self.rm_pkg, "-y", *self.salt_pkgs)
            self._check_retcode(ret)

    def write_launchd_conf(self, service):
        service_name = f"com.saltstack.salt.{service}"
        ret = self.proc.run("launchctl", "list", service_name)
        # 113 means it couldn't find a service with that name
        if ret.returncode == 113:
            daemons_dir = pathlib.Path("/Library", "LaunchDaemons")
            plist_file = daemons_dir / f"{service_name}.plist"
            # Make sure we're using this plist file
            if plist_file.exists():
                log.warning("Removing existing plist file for service: %s", service)
                plist_file.unlink()

            log.debug("Creating plist file for service: %s", service)
            contents = f"""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
                <plist version="1.0">
                    <dict>
                        <key>Label</key>
                        <string>{service_name}</string>
                        <key>RunAtLoad</key>
                        <true/>
                        <key>KeepAlive</key>
                        <true/>
                        <key>ProgramArguments</key>
                        <array>"""
            for part in self.binary_paths[service]:
                contents += (
                    f"""\n                            <string>{part}</string>\n"""
                )
            contents += f"""\
                            <string>-c</string>
                            <string>{self.conf_dir}</string>
                        </array>
                        <key>SoftResourceLimits</key>
                        <dict>
                            <key>NumberOfFiles</key>
                            <integer>100000</integer>
                        </dict>
                        <key>HardResourceLimits</key>
                        <dict>
                            <key>NumberOfFiles</key>
                            <integer>100000</integer>
                        </dict>
                    </dict>
                </plist>
                """
            plist_file.write_text(textwrap.dedent(contents), encoding="utf-8")
            contents = plist_file.read_text()
            log.debug("Created '%s'. Contents:\n%s", plist_file, contents)

            # Delete the plist file upon completion
            atexit.register(plist_file.unlink)

    def write_systemd_conf(self, service, binary):
        ret = self.proc.run("systemctl", "daemon-reload")
        self._check_retcode(ret)
        ret = self.proc.run("systemctl", "status", service)
        if ret.returncode == 4:
            log.warning(
                "No systemd unit file was found for service %s. Creating one.", service
            )
            contents = textwrap.dedent(
                """\
                [Unit]
                Description={service}

                [Service]
                KillMode=process
                Type=notify
                NotifyAccess=all
                LimitNOFILE=8192
                ExecStart={tgt} -c {conf_dir}

                [Install]
                WantedBy=multi-user.target
                """
            )
            if isinstance(binary, list) and len(binary) == 1:
                binary = shutil.which(binary[0]) or binary[0]
            elif isinstance(binary, list):
                binary = " ".join(binary)
            unit_path = pathlib.Path(f"/etc/systemd/system/{service}.service")
            contents = contents.format(
                service=service, tgt=binary, conf_dir=self.conf_dir
            )
            log.info("Created '%s'. Contents:\n%s", unit_path, contents)
            unit_path.write_text(contents, encoding="utf-8")
            ret = self.proc.run("systemctl", "daemon-reload")
            atexit.register(unit_path.unlink)
            self._check_retcode(ret)

    def __enter__(self):
        if platform.is_windows():
            self.update_process_path()

        if self.no_install:
            return self

        if self.upgrade:
            self.install_previous()
        else:
            # assume downgrade, since no_install only used in these two cases
            self.install()
        return self

    def __exit__(self, *_):
        if not self.no_uninstall:
            self.uninstall()

        # Did we left anything running?!
        procs = []
        for proc in psutil.process_iter():
            if "salt" in proc.name():
                cmdl_strg = " ".join(str(element) for element in _get_cmdline(proc))
                if "/opt/saltstack" in cmdl_strg:
                    procs.append(proc)

        if procs:
            terminate_process_list(procs, kill=True, slow_stop=True)


class PkgSystemdSaltDaemonImpl(SystemdSaltDaemonImpl):
    # pylint: disable=access-member-before-definition
    def get_service_name(self):
        if self._service_name is None:
            self._service_name = self.factory.script_name
        return self._service_name

    # pylint: enable=access-member-before-definition


@attr.s(kw_only=True)
class PkgLaunchdSaltDaemonImpl(PkgSystemdSaltDaemonImpl):

    plist_file = attr.ib()

    @plist_file.default
    def _default_plist_file(self):
        daemons_dir = pathlib.Path("/Library", "LaunchDaemons")
        return daemons_dir / f"{self.get_service_name()}.plist"

    def get_service_name(self):
        if self._service_name is None:
            service_name = super().get_service_name()
            if "-" in service_name:
                service_name = service_name.split("-")[-1]
            self._service_name = f"com.saltstack.salt.{service_name}"
        return self._service_name

    def cmdline(self, *args):  # pylint: disable=arguments-differ
        """
        Construct a list of arguments to use when starting the subprocess.

        :param str args:
            Additional arguments to use when starting the subprocess

        """
        if args:  # pragma: no cover
            log.debug(
                "%s.run() is ignoring the passed in arguments: %r",
                self.__class__.__name__,
                args,
            )
        self._internal_run(
            "launchctl",
            "enable",
            f"system/{self.get_service_name()}",
        )
        return (
            "launchctl",
            "bootstrap",
            "system",
            str(self.plist_file),
        )

    def is_running(self):
        """
        Returns true if the sub-process is alive.
        """
        if self._process is None:
            ret = self._internal_run("launchctl", "list", self.get_service_name())
            if ret.stdout == "":
                return False

            if "PID" not in ret.stdout:
                return False

            pid = None
            # PID in a line that looks like this
            # "PID" = 445;
            for line in ret.stdout.splitlines():
                if "PID" in line:
                    pid = line.rstrip(";").split(" = ")[1]

            if pid is None:
                return False

            self._process = psutil.Process(int(pid))

        return self._process.is_running()

    def _terminate(self):
        """
        This method actually terminates the started daemon.
        """
        # We completely override the parent class method because we're not using
        # the self._terminal property, it's a launchd service
        if self._process is None:  # pragma: no cover
            # pylint: disable=access-member-before-definition
            if TYPE_CHECKING:
                # Make mypy happy
                assert self._terminal_result
            return self._terminal_result
            # pylint: enable=access-member-before-definition

        atexit.unregister(self.terminate)
        log.info("Stopping %s", self.factory)
        pid = self.pid
        # Collect any child processes information before terminating the process
        with contextlib.suppress(psutil.NoSuchProcess):
            for child in psutil.Process(pid).children(recursive=True):
                # pylint: disable=access-member-before-definition
                if child not in self._children:
                    self._children.append(child)
                # pylint: enable=access-member-before-definition

        if self._process.is_running():  # pragma: no cover
            cmdline = _get_cmdline(self._process)
        else:
            cmdline = []

        # Disable the service
        self._internal_run(
            "launchctl",
            "disable",
            f"system/{self.get_service_name()}",
        )
        # Unload the service
        self._internal_run("launchctl", "bootout", "system", str(self.plist_file))

        if self._process.is_running():  # pragma: no cover
            try:
                self._process.wait()
            except psutil.TimeoutExpired:
                self._process.terminate()
                try:
                    self._process.wait()
                except psutil.TimeoutExpired:
                    pass

        exitcode = self._process.wait() or 0

        # Dereference the internal _process attribute
        self._process = None
        # Lets log and kill any child processes left behind, including the main subprocess
        # if it failed to properly stop
        terminate_process(
            pid=pid,
            kill_children=True,
            children=self._children,  # pylint: disable=access-member-before-definition
            slow_stop=self.factory.slow_stop,
        )

        # pylint: disable=access-member-before-definition
        if self._terminal_stdout is not None:
            self._terminal_stdout.close()
        if self._terminal_stderr is not None:
            self._terminal_stderr.close()
        # pylint: enable=access-member-before-definition
        stdout = stderr = ""
        try:
            self._terminal_result = ProcessResult(
                returncode=exitcode, stdout=stdout, stderr=stderr, cmdline=cmdline
            )
            log.info("%s %s", self.factory.__class__.__name__, self._terminal_result)
            return self._terminal_result
        finally:
            self._terminal = None
            self._terminal_stdout = None
            self._terminal_stderr = None
            self._terminal_timeout = None
            self._children = []


@attr.s(kw_only=True)
class PkgSsmSaltDaemonImpl(PkgSystemdSaltDaemonImpl):
    def cmdline(self, *args):  # pylint: disable=arguments-differ
        """
        Construct a list of arguments to use when starting the subprocess.

        :param str args:
            Additional arguments to use when starting the subprocess

        """
        if args:  # pragma: no cover
            log.debug(
                "%s.run() is ignoring the passed in arguments: %r",
                self.__class__.__name__,
                args,
            )
        return (
            str(self.factory.salt_pkg_install.ssm_bin),
            "start",
            self.get_service_name(),
        )

    def is_running(self):
        """
        Returns true if the sub-process is alive.
        """
        if self._process is None:
            n = 1
            while True:
                if self._process is not None:
                    break
                time.sleep(1)
                ret = self._internal_run(
                    str(self.factory.salt_pkg_install.ssm_bin),
                    "processes",
                    self.get_service_name(),
                )
                log.debug("process result %s", ret)
                if not ret.stdout or (ret.stdout and not ret.stdout.strip()):
                    if n >= 120:
                        return False
                    n += 1
                    continue
                for line in ret.stdout.splitlines():
                    log.warning("Line: %s", line)
                    if not line.strip():
                        continue
                    mainpid = line.strip().split()[0]
                    self._process = psutil.Process(int(mainpid))
                    break
        ret = self._process.is_running()
        if not hasattr(self, "logged_running"):
            log.error("SSM processs is running %s", ret)
            self.logged_running = True
        return ret

    def _terminate(self):
        """
        This method actually terminates the started daemon.
        """
        # We completely override the parent class method because we're not using the
        # self._terminal property, it's a systemd service
        if self._process is None:  # pragma: no cover
            # pylint: disable=access-member-before-definition
            if TYPE_CHECKING:
                # Make mypy happy
                assert self._terminal_result
            return self._terminal_result
            # pylint: enable=access-member-before-definition

        atexit.unregister(self.terminate)
        log.info("Stopping %s", self.factory)
        pid = self.pid
        # Collect any child processes information before terminating the process
        with contextlib.suppress(psutil.NoSuchProcess):
            for child in psutil.Process(pid).children(recursive=True):
                # pylint: disable=access-member-before-definition
                if child not in self._children:
                    self._children.append(child)
                # pylint: enable=access-member-before-definition

        if self._process.is_running():  # pragma: no cover
            cmdline = _get_cmdline(self._process)
        else:
            cmdline = []

        # Tell ssm to stop the service
        try:
            self._internal_run(
                str(self.factory.salt_pkg_install.ssm_bin),
                "stop",
                self.get_service_name(),
            )
        except FileNotFoundError:
            pass

        if self._process.is_running():  # pragma: no cover
            try:
                self._process.wait()
            except psutil.TimeoutExpired:
                self._process.terminate()
                try:
                    self._process.wait()
                except psutil.TimeoutExpired:
                    pass

        exitcode = self._process.wait() or 0

        # Dereference the internal _process attribute
        self._process = None
        # Let's log and kill any child processes left behind, including the main
        # subprocess if it failed to properly stop
        terminate_process(
            pid=pid,
            kill_children=True,
            children=self._children,  # pylint: disable=access-member-before-definition
            slow_stop=self.factory.slow_stop,
        )

        # pylint: disable=access-member-before-definition
        if self._terminal_stdout is not None:
            self._terminal_stdout.close()
        if self._terminal_stderr is not None:
            self._terminal_stderr.close()
        # pylint: enable=access-member-before-definition
        stdout = stderr = ""
        try:
            self._terminal_result = ProcessResult(
                returncode=exitcode, stdout=stdout, stderr=stderr, cmdline=cmdline
            )
            log.info("%s %s", self.factory.__class__.__name__, self._terminal_result)
            return self._terminal_result
        finally:
            self._terminal = None
            self._terminal_stdout = None
            self._terminal_stderr = None
            self._terminal_timeout = None
            self._children = []


@attr.s(kw_only=True)
class PkgMixin:
    salt_pkg_install: SaltPkgInstall = attr.ib()

    def get_script_path(self):
        if platform.is_darwin() and self.salt_pkg_install.classic:
            if self.salt_pkg_install.run_root and os.path.exists(
                self.salt_pkg_install.run_root
            ):
                return str(self.salt_pkg_install.run_root)
            elif os.path.exists(self.salt_pkg_install.bin_dir / self.script_name):
                return str(self.salt_pkg_install.bin_dir / self.script_name)
            else:
                return str(self.salt_pkg_install.install_dir / self.script_name)
        return super().get_script_path()

    def cmdline(self, *args, **kwargs):
        _cmdline = super().cmdline(*args, **kwargs)
        if _cmdline[0] == self.python_executable:
            _cmdline.pop(0)
        return _cmdline


@attr.s(kw_only=True)
class DaemonPkgMixin(PkgMixin):
    def __attrs_post_init__(self):
        if not platform.is_windows() and self.salt_pkg_install.pkg_system_service:
            if platform.is_darwin():
                self.write_launchd_conf()
            else:
                self.write_systemd_conf()

    def get_service_name(self):
        return self.script_name

    def write_launchd_conf(self):
        raise NotImplementedError

    def write_systemd_conf(self):
        raise NotImplementedError


@attr.s(kw_only=True)
class SaltMaster(DaemonPkgMixin, master.SaltMaster):
    """
    Subclassed just to tweak the binary paths if needed and factory classes.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt-master"
        master.SaltMaster.__attrs_post_init__(self)
        DaemonPkgMixin.__attrs_post_init__(self)

    def _get_impl_class(self):
        if self.system_service and self.salt_pkg_install.pkg_system_service:
            if platform.is_windows():
                return PkgSsmSaltDaemonImpl
            if platform.is_darwin():
                return PkgLaunchdSaltDaemonImpl
            return PkgSystemdSaltDaemonImpl
        return DaemonImpl

    def write_launchd_conf(self):
        self.salt_pkg_install.write_launchd_conf("master")

    def write_systemd_conf(self):
        self.salt_pkg_install.write_systemd_conf(
            "salt-master", self.salt_pkg_install.binary_paths["master"]
        )

    def salt_minion_daemon(self, minion_id, **kwargs):
        return super().salt_minion_daemon(
            minion_id,
            factory_class=SaltMinion,
            salt_pkg_install=self.salt_pkg_install,
            **kwargs,
        )

    def salt_api_daemon(self, **kwargs):
        return super().salt_api_daemon(
            factory_class=SaltApi, salt_pkg_install=self.salt_pkg_install, **kwargs
        )

    def salt_key_cli(self, factory_class=None, **factory_class_kwargs):
        if not factory_class:
            factory_class = SaltKey
        factory_class_kwargs["salt_pkg_install"] = self.salt_pkg_install
        return super().salt_key_cli(
            factory_class=factory_class,
            **factory_class_kwargs,
        )

    def salt_cli(self, factory_class=None, **factory_class_kwargs):
        if not factory_class:
            factory_class = SaltCli
        factory_class_kwargs["salt_pkg_install"] = self.salt_pkg_install
        return super().salt_cli(
            factory_class=factory_class,
            **factory_class_kwargs,
        )


@attr.s(kw_only=True)
class SaltMasterWindows(SaltMaster):
    """
    Subclassed just to tweak the binary paths if needed and factory classes.
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.script_name = cli_scripts.generate_script(
            bin_dir=self.factories_manager.scripts_dir,
            script_name="salt-master",
            code_dir=self.factories_manager.code_dir.parent,
        )

        # XXX: Add install path to cli_scripts.generate_scripts?
        def patch_script(script):
            text = script.read_text()
            newlines = []
            for line in text.splitlines():
                newlines.append(line)
                if line == "sys.path.insert(0, CODE_DIR)":
                    newlines.append(
                        'sys.path.insert(0, "C:\\Program Files\\Salt Project\\Salt\\Lib\\site-packages")'
                    )
            script.write_text(os.linesep.join(newlines))

        patch_script(self.factories_manager.scripts_dir / "cli_salt_master.py")

    def _get_impl_class(self):
        return DaemonImpl

    def cmdline(self, *args, **kwargs):
        cmdline_ = super().cmdline(*args, **kwargs)
        if self.python_executable:
            if cmdline_[0] != self.python_executable:
                cmdline_.insert(0, self.python_executable)
        return cmdline_


@attr.s(kw_only=True, slots=True)
class SaltMinion(DaemonPkgMixin, minion.SaltMinion):
    """
    Subclassed just to tweak the binary paths if needed and factory classes.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt-minion"
        minion.SaltMinion.__attrs_post_init__(self)
        DaemonPkgMixin.__attrs_post_init__(self)

    def _get_impl_class(self):
        if self.system_service and self.salt_pkg_install.pkg_system_service:
            if platform.is_windows():
                return PkgSsmSaltDaemonImpl
            if platform.is_darwin():
                return PkgLaunchdSaltDaemonImpl
            return PkgSystemdSaltDaemonImpl
        return DaemonImpl

    def write_launchd_conf(self):
        self.salt_pkg_install.write_launchd_conf("minion")

    def write_systemd_conf(self):
        self.salt_pkg_install.write_systemd_conf(
            "salt-minion", self.salt_pkg_install.binary_paths["minion"]
        )

    def salt_call_cli(self, factory_class=None, **factory_class_kwargs):
        if not factory_class:
            factory_class = SaltCall
        factory_class_kwargs["salt_pkg_install"] = self.salt_pkg_install
        return super().salt_call_cli(
            factory_class=factory_class,
            **factory_class_kwargs,
        )


@attr.s(kw_only=True, slots=True)
class SaltApi(DaemonPkgMixin, api.SaltApi):
    """
    Subclassed just to tweak the binary paths if needed.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt-api"
        api.SaltApi.__attrs_post_init__(self)
        DaemonPkgMixin.__attrs_post_init__(self)

    def _get_impl_class(self):
        if self.system_service and self.salt_pkg_install.pkg_system_service:
            if platform.is_windows():
                return PkgSsmSaltDaemonImpl
            if platform.is_darwin():
                return PkgLaunchdSaltDaemonImpl
            return PkgSystemdSaltDaemonImpl
        return DaemonImpl

    def write_launchd_conf(self):
        self.salt_pkg_install.write_launchd_conf("api")

    def write_systemd_conf(self):
        self.salt_pkg_install.write_systemd_conf(
            "salt-api",
            self.salt_pkg_install.binary_paths["api"],
        )


@attr.s(kw_only=True, slots=True)
class SaltCall(PkgMixin, call.SaltCall):
    """
    Subclassed just to tweak the binary paths if needed.
    """

    def __attrs_post_init__(self):
        call.SaltCall.__attrs_post_init__(self)
        self.script_name = "salt-call"


@attr.s(kw_only=True, slots=True)
class SaltCli(PkgMixin, saltfactories.cli.salt.SaltCli):
    """
    Subclassed just to tweak the binary paths if needed.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt"
        saltfactories.cli.salt.SaltCli.__attrs_post_init__(self)


@attr.s(kw_only=True, slots=True)
class SaltKey(PkgMixin, key.SaltKey):
    """
    Subclassed just to tweak the binary paths if needed.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt-key"
        key.SaltKey.__attrs_post_init__(self)


@attr.s(kw_only=True, slots=True)
class ApiRequest:
    port: int = attr.ib(repr=False)
    account: TestAccount = attr.ib(repr=False)
    session: requests.Session = attr.ib(init=False, repr=False)
    api_uri: str = attr.ib(init=False)
    auth_data: dict[str, str] = attr.ib(init=False)

    @session.default
    def _default_session(self):
        return requests.Session()

    @api_uri.default
    def _default_api_uri(self):
        return f"http://localhost:{self.port}"

    @auth_data.default
    def _default_auth_data(self):
        return {
            "username": self.account.username,
            "password": self.account.password,
            "eauth": "auto",
            "out": "json",
        }

    def post(self, url, data):
        post_data = dict(**self.auth_data, **data)
        resp = self.session.post(f"{self.api_uri}/run", data=post_data).json()
        minion = next(iter(resp["return"][0]))
        return resp["return"][0][minion]

    def __enter__(self):
        self.session.__enter__()
        return self

    def __exit__(self, *args):
        self.session.__exit__(*args)



import pyrax  # pylint: disable=3rd-party-module-not-gated


class Authenticate:
    def __init__(self, username, password, region, **kwargs):
        cloud_kwargs = {"password": password, "region": region}
        pyrax.settings.set("identity_type", kwargs.get("identity_type", "rackspace"))
        pyrax.settings.set(
            "auth_endpoint",
            kwargs.get("auth_endpoint", "https://identity.api.rackspacecloud.com/v2.0"),
        )

        pyrax.set_credentials(username, **cloud_kwargs)
        self.conn = pyrax



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest

import salt.states.http as http
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {http: {}}


def test_query():
    """
    Test to perform an HTTP query and statefully return the result
    """
    ret = [
        {
            "changes": {},
            "comment": (
                " Either match text (match) or a status code (status) is required."
            ),
            "data": {},
            "name": "salt",
            "result": False,
        },
        {
            "changes": {},
            "comment": " (TEST MODE)",
            "data": True,
            "name": "salt",
            "result": None,
        },
    ]
    assert http.query("salt") == ret[0]

    with patch.dict(http.__opts__, {"test": True}):
        mock = MagicMock(return_value=True)
        with patch.dict(http.__salt__, {"http.query": mock}):
            assert http.query("salt", "Dude", "stack") == ret[1]

    with patch.dict(http.__opts__, {"test": False}):
        mock = MagicMock(return_value={"body": "http body", "status": 200})
        expected = {
            "name": "http://example.com/",
            "result": True,
            "comment": "Status 200 was found.",
            "changes": {},
            "data": {"body": "http body", "status": 200},
        }

        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(name="http://example.com/", status=200, decode=False)
                == expected
            )

    with patch.dict(http.__opts__, {"test": False}):
        mock = MagicMock(return_value={"body": "http body", "status": 200})
        expected = {
            "name": "http://example.com/",
            "result": True,
            "comment": "Status 200 was found.",
            "changes": {},
            "data": {"body": "http body", "status": 200},
        }

        with patch.dict(http.__salt__, {"http.wait_for_successful_query": mock}):
            assert (
                http.query(name="http://example.com/", status=200, wait_for=300)
                == expected
            )

    with patch.dict(http.__opts__, {"test": True}):
        mock = MagicMock(return_value={"body": "http body", "status": 200})
        expected = {
            "name": "http://example.com/",
            "result": None,
            "comment": "Status 200 was found. (TEST MODE, TEST URL WAS: http://status.example.com)",
            "changes": {},
            "data": {"body": "http body", "status": 200},
        }

        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(
                    name="http://example.com/",
                    status=200,
                    test_url="http://status.example.com",
                )
                == expected
            )


def test_query_pcre_statustype():
    """
    Test to perform an HTTP query with a regex used to match the status code and statefully return the result
    """
    testurl = "salturl"

    with patch.dict(http.__opts__, {"test": False}):
        http_result = {"text": "This page returned a 201 status code", "status": "201"}
        mock = MagicMock(return_value=http_result)

        state_return = {
            "changes": {},
            "comment": (
                'Match text "This page returned" was found. Status pattern "200|201" was'
                " found."
            ),
            "data": {"status": "201", "text": "This page returned a 201 status code"},
            "name": testurl,
            "result": True,
        }

        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(
                    testurl,
                    match="This page returned",
                    status="200|201",
                    status_type="pcre",
                )
                == state_return
            )

    with patch.dict(http.__opts__, {"test": False}):
        http_result = {"text": "This page returned a 201 status code", "status": "201"}
        mock = MagicMock(return_value=http_result)

        state_return = {
            "changes": {},
            "comment": ('Status pattern "200|201" was found.'),
            "data": {"status": "201", "text": "This page returned a 201 status code"},
            "name": testurl,
            "result": True,
        }

        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(
                    testurl,
                    status="200|201",
                    status_type="pcre",
                )
                == state_return
            )

        http_result = {"text": "This page returned a 403 status code", "status": "403"}
        mock = MagicMock(return_value=http_result)

        state_return = {
            "name": "salturl",
            "result": False,
            "comment": 'Match text "This page returned" was found. Status pattern "200|201" was not found.',
            "changes": {},
            "data": {"text": "This page returned a 403 status code", "status": "403"},
        }

        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(
                    testurl,
                    match="This page returned",
                    status="200|201",
                    status_type="pcre",
                )
                == state_return
            )


def test_query_pcre_matchtype():
    """
    Test to perform an HTTP query with a regex used to match the returned text and statefully return the result
    """
    testurl = "salturl"

    with patch.dict(http.__opts__, {"test": False}):
        http_result = {"text": "This page returned a 201 status code", "status": "201"}
        mock = MagicMock(return_value=http_result)

        state_return = {
            "changes": {},
            "comment": ('Match pattern "This page returned" was found.'),
            "data": {"status": "201", "text": "This page returned a 201 status code"},
            "name": testurl,
            "result": True,
        }

        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(
                    testurl,
                    match="This page returned",
                    match_type="pcre",
                )
                == state_return
            )

        http_result = {
            "text": "This page did not return a 201 status code",
            "status": "403",
        }
        mock = MagicMock(return_value=http_result)

        state_return = {
            "changes": {},
            "comment": ('Match pattern "This page returned" was not found.'),
            "data": {
                "status": "403",
                "text": "This page did not return a 201 status code",
            },
            "name": testurl,
            "result": False,
        }

        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(
                    testurl,
                    match="This page returned",
                    match_type="pcre",
                )
                == state_return
            )


def test_query_stringstatustype():
    """
    Test to perform an HTTP query with a string status code and statefully return the result
    """
    testurl = "salturl"

    with patch.dict(http.__opts__, {"test": False}):
        http_result = {"text": "This page returned a 201 status code", "status": "201"}
        mock = MagicMock(return_value=http_result)

        with patch.dict(http.__salt__, {"http.query": mock}):
            state_return = {
                "changes": {},
                "comment": 'Match text "This page returned" was found. Status 201 was found.',
                "data": {
                    "status": "201",
                    "text": "This page returned a 201 status code",
                },
                "name": testurl,
                "result": True,
            }

            assert (
                http.query(
                    testurl,
                    match="This page returned",
                    status="201",
                    status_type="string",
                )
                == state_return
            )

        http_result = {"text": "This page returned a 403 status code", "status": "403"}
        mock = MagicMock(return_value=http_result)

        with patch.dict(http.__salt__, {"http.query": mock}):
            state_return = {
                "name": "salturl",
                "result": False,
                "comment": 'Match text "This page returned" was found. Status 201 was not found.',
                "changes": {},
                "data": {
                    "text": "This page returned a 403 status code",
                    "status": "403",
                },
            }

            assert (
                http.query(
                    testurl,
                    match="This page returned",
                    status="201",
                    status_type="string",
                )
                == state_return
            )


def test_query_invalidstatustype():
    """
    Test to perform an HTTP query with a string status code and statefully return the result
    """
    testurl = "salturl"

    with patch.dict(http.__opts__, {"test": False}):
        http_result = {"text": "This page returned a 201 status code", "status": "201"}
        mock = MagicMock(return_value=http_result)

        with patch.dict(http.__salt__, {"http.query": mock}):
            state_return = {
                "name": "salturl",
                "result": None,
                "comment": "",
                "changes": {},
                "data": {
                    "text": "This page returned a 201 status code",
                    "status": "201",
                },
            }

            assert (
                http.query(
                    testurl,
                    status="201",
                    status_type="invalid",
                )
                == state_return
            )


def test_query_liststatustype():
    """
    Test to perform an HTTP query with a list of status codes and statefully return the result
    """
    testurl = "salturl"

    with patch.dict(http.__opts__, {"test": False}):
        http_result = {"text": "This page returned a 201 status code", "status": "201"}
        state_return = {
            "changes": {},
            "comment": 'Match text "This page returned" was found. Status 201 was found.',
            "data": {"status": "201", "text": "This page returned a 201 status code"},
            "name": testurl,
            "result": True,
        }

        mock = MagicMock(return_value=http_result)
        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(
                    testurl,
                    match="This page returned",
                    status=["200", "201"],
                    status_type="list",
                )
                == state_return
            )

    with patch.dict(http.__opts__, {"test": False}):
        http_result = {"text": "This page returned a 201 status code", "status": "201"}
        state_return = {
            "changes": {},
            "comment": "Status 201 was found.",
            "data": {"status": "201", "text": "This page returned a 201 status code"},
            "name": testurl,
            "result": True,
        }

        mock = MagicMock(return_value=http_result)
        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(
                    testurl,
                    status=["200", "201"],
                    status_type="list",
                )
                == state_return
            )

        http_result = {"text": "This page returned a 403 status code", "status": "403"}
        state_return = {
            "name": "salturl",
            "result": False,
            "comment": "Match text \"This page returned a 200\" was not found. Statuses ['200', '201'] were not found.",
            "changes": {},
            "data": {"text": "This page returned a 403 status code", "status": "403"},
        }

        mock = MagicMock(return_value=http_result)
        with patch.dict(http.__salt__, {"http.query": mock}):
            assert (
                http.query(
                    testurl,
                    match="This page returned a 200",
                    status=["200", "201"],
                    status_type="list",
                )
                == state_return
            )


def test_wait_for_with_interval():
    """
    Test for wait_for_successful_query waits for request_interval
    """

    query_mock = MagicMock(side_effect=[{"error": "error"}, {"result": True}])

    with patch.object(http, "query", query_mock):
        with patch("time.sleep", MagicMock()) as sleep_mock:
            assert http.wait_for_successful_query(
                "url", request_interval=1, status=200
            ) == {"result": True}
            sleep_mock.assert_called_once_with(1)


def test_wait_for_without_interval():
    """
    Test for wait_for_successful_query waits for request_interval
    """

    query_mock = MagicMock(side_effect=[{"error": "error"}, {"result": True}])

    with patch.object(http, "query", query_mock):
        with patch("time.sleep", MagicMock()) as sleep_mock:
            assert http.wait_for_successful_query("url", status=200) == {"result": True}
            sleep_mock.assert_not_called()

    query_mock = MagicMock(return_value={"result": False})

    with patch.object(http, "query", query_mock):
        with patch(
            "time.time", MagicMock(side_effect=[1697564521.9640958, 1697564822.9640958])
        ):
            assert http.wait_for_successful_query("url", status=200) == {
                "result": False
            }

    query_mock = MagicMock(side_effect=Exception())

    with patch.object(http, "query", query_mock):
        with patch(
            "time.time", MagicMock(side_effect=[1697564521.9640958, 1697564822.9640958])
        ):
            with pytest.raises(Exception):
                http.wait_for_successful_query("url", status=200)



import os
from subprocess import PIPE
from textwrap import dedent

import pytest

import salt.renderers.gpg as gpg
from salt.exceptions import SaltRenderError
from tests.support.mock import MagicMock, Mock, call, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    """
    GPG renderer configuration
    """
    minion_opts["gpg_decrypt_must_succeed"] = True
    return {gpg: {"__opts__": minion_opts}}


def test__get_gpg_exec():
    """
    test _get_gpg_exec
    """
    gpg_exec = "/bin/gpg"

    with patch("salt.utils.path.which", MagicMock(return_value=gpg_exec)):
        assert gpg._get_gpg_exec() == gpg_exec

    with patch("salt.utils.path.which", MagicMock(return_value=False)):
        pytest.raises(SaltRenderError, gpg._get_gpg_exec)


def test__decrypt_ciphertext():
    """
    test _decrypt_ciphertext
    """
    key_dir = "/etc/salt/gpgkeys"
    secret = "Use more salt."
    crypted = "-----BEGIN PGP MESSAGE-----!@#$%^&*()_+-----END PGP MESSAGE-----"

    multisecret = "password is {0} and salt is {0}".format(secret)
    multicrypted = "password is {0} and salt is {0}".format(crypted)

    class GPGDecrypt:
        def communicate(self, *args, **kwargs):
            return [secret, None]

    class GPGNotDecrypt:
        def communicate(self, *args, **kwargs):
            return [None, "decrypt error"]

    with patch(
        "salt.renderers.gpg._get_key_dir", MagicMock(return_value=key_dir)
    ), patch("salt.utils.path.which", MagicMock()):
        with patch("salt.renderers.gpg.Popen", MagicMock(return_value=GPGDecrypt())):
            assert gpg._decrypt_ciphertexts(crypted) == secret
            assert gpg._decrypt_ciphertexts(multicrypted) == multisecret
        with patch("salt.renderers.gpg.Popen", MagicMock(return_value=GPGNotDecrypt())):
            with pytest.raises(SaltRenderError) as decrypt_error:
                gpg._decrypt_ciphertexts(crypted)
            # Assertions must be made after closure of context manager
            assert decrypt_error.value.args[0].startswith("Could not decrypt cipher ")
            assert crypted in decrypt_error.value.args[0]
            assert "decrypt error" in decrypt_error.value.args[0]
            with pytest.raises(SaltRenderError) as multidecrypt_error:
                gpg._decrypt_ciphertexts(multicrypted)
            assert multidecrypt_error.value.args[0].startswith(
                "Could not decrypt cipher "
            )
            # Function will raise on a single ciphertext even if multiple are passed
            assert crypted in multidecrypt_error.value.args[0]
            assert "decrypt error" in multidecrypt_error.value.args[0]


def test__decrypt_object():
    """
    test _decrypt_object
    """
    secret = "Use more salt."
    crypted = "-----BEGIN PGP MESSAGE-----!@#$%^&*()_+-----END PGP MESSAGE-----"

    secret_map = {"secret": secret}
    crypted_map = {"secret": crypted}

    secret_list = [secret]
    crypted_list = [crypted]

    with patch(
        "salt.renderers.gpg._decrypt_ciphertext", MagicMock(return_value=secret)
    ):
        assert gpg._decrypt_object(secret) == secret
        assert gpg._decrypt_object(crypted) == secret
        assert gpg._decrypt_object(crypted_map) == secret_map
        assert gpg._decrypt_object(crypted_list) == secret_list
        assert gpg._decrypt_object(None) is None


def test_render():
    """
    test render
    """
    key_dir = "/etc/salt/gpgkeys"
    secret = "Use more salt."
    crypted = "-----BEGIN PGP MESSAGE-----!@#$%^&*()_+"

    with patch("salt.renderers.gpg._get_gpg_exec", MagicMock(return_value=True)):
        with patch("salt.renderers.gpg._get_key_dir", MagicMock(return_value=key_dir)):
            with patch(
                "salt.renderers.gpg._decrypt_object", MagicMock(return_value=secret)
            ):
                assert gpg.render(crypted) == secret


def test_render_bytes():
    """
    test rendering bytes
    """
    key_dir = "/etc/salt/gpgkeys"
    binfo = b"User more salt."

    with patch("salt.renderers.gpg._get_gpg_exec", MagicMock(return_value=True)):
        with patch("salt.renderers.gpg._get_key_dir", MagicMock(return_value=key_dir)):
            assert gpg.render(binfo) == binfo.decode()


def test_multi_render():
    key_dir = "/etc/salt/gpgkeys"
    secret = "Use more salt."
    expected = "\n".join([secret] * 3)
    crypted = dedent(
        """\
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
    """
    )

    with patch("salt.renderers.gpg._get_gpg_exec", MagicMock(return_value=True)):
        with patch("salt.renderers.gpg._get_key_dir", MagicMock(return_value=key_dir)):
            with patch(
                "salt.renderers.gpg._decrypt_ciphertext",
                MagicMock(return_value=secret),
            ):
                assert gpg.render(crypted) == expected


def test_render_with_binary_data_should_return_binary_data():
    key_dir = "/etc/salt/gpgkeys"
    secret = b"Use\x8b more\x8b salt."
    expected = b"\n".join([secret] * 3)
    crypted = dedent(
        """\
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
    """
    )

    with patch("salt.renderers.gpg._get_gpg_exec", MagicMock(return_value=True)):
        with patch("salt.renderers.gpg._get_key_dir", MagicMock(return_value=key_dir)):
            with patch(
                "salt.renderers.gpg._decrypt_ciphertext",
                MagicMock(return_value=secret),
            ):
                assert gpg.render(crypted, encoding="utf-8") == expected


def test_render_with_translate_newlines_should_translate_newlines():
    key_dir = "/etc/salt/gpgkeys"
    secret = b"Use\x8b more\x8b salt."
    expected = b"\n\n".join([secret] * 3)
    crypted = dedent(
        """\
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----\\n
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----\\n
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
    """
    )

    with patch("salt.renderers.gpg._get_gpg_exec", MagicMock(return_value=True)):
        with patch("salt.renderers.gpg._get_key_dir", MagicMock(return_value=key_dir)):
            with patch(
                "salt.renderers.gpg._decrypt_ciphertext",
                MagicMock(return_value=secret),
            ):
                assert (
                    gpg.render(crypted, translate_newlines=True, encoding="utf-8")
                    == expected
                )


def test_render_without_cache():
    key_dir = "/etc/salt/gpgkeys"
    secret = "Use more salt."
    expected = "\n".join([secret] * 3)
    crypted = dedent(
        """\
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
    """
    )

    with patch("salt.renderers.gpg.Popen") as popen_mock:
        popen_mock.return_value = Mock(
            communicate=lambda *args, **kwargs: (secret, None),
        )
        with patch(
            "salt.renderers.gpg._get_gpg_exec",
            MagicMock(return_value="/usr/bin/gpg"),
        ):
            with patch(
                "salt.renderers.gpg._get_key_dir", MagicMock(return_value=key_dir)
            ):
                assert gpg.render(crypted) == expected
                gpg_call = call(
                    [
                        "/usr/bin/gpg",
                        "--homedir",
                        "/etc/salt/gpgkeys",
                        "--status-fd",
                        "2",
                        "--no-tty",
                        "-d",
                    ],
                    shell=False,
                    stderr=PIPE,
                    stdin=PIPE,
                    stdout=PIPE,
                )
                popen_mock.assert_has_calls([gpg_call] * 3)


def test_render_with_cache(minion_opts):
    key_dir = "/etc/salt/gpgkeys"
    secret = "Use more salt."
    expected = "\n".join([secret] * 3)
    crypted = dedent(
        """\
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
        -----BEGIN PGP MESSAGE-----
        !@#$%^&*()_+
        -----END PGP MESSAGE-----
    """
    )

    minion_opts["gpg_cache"] = True
    with patch.dict(gpg.__opts__, minion_opts):
        with patch("salt.renderers.gpg.Popen") as popen_mock:
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwargs: (secret, None),
            )
            with patch(
                "salt.renderers.gpg._get_gpg_exec",
                MagicMock(return_value="/usr/bin/gpg"),
            ):
                with patch(
                    "salt.renderers.gpg._get_key_dir",
                    MagicMock(return_value=key_dir),
                ):
                    with patch(
                        "salt.utils.atomicfile.atomic_open",
                        MagicMock(),
                    ) as atomic_open_mock:
                        assert gpg.render(crypted) == expected
                        gpg_call = call(
                            [
                                "/usr/bin/gpg",
                                "--homedir",
                                "/etc/salt/gpgkeys",
                                "--status-fd",
                                "2",
                                "--no-tty",
                                "-d",
                            ],
                            shell=False,
                            stderr=PIPE,
                            stdin=PIPE,
                            stdout=PIPE,
                        )
                        popen_mock.assert_has_calls([gpg_call] * 1)
                        atomic_open_mock.assert_has_calls(
                            [
                                call(
                                    os.path.join(minion_opts["cachedir"], "gpg_cache"),
                                    "wb+",
                                )
                            ]
                        )



"""
Manage Autoscale Groups
=======================

.. versionadded:: 2014.7.0

Create and destroy autoscale groups. Be aware that this interacts with Amazon's
services, and so may incur charges.

This module uses boto, which can be installed via package, or pip.

This module accepts explicit autoscale credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More Information available at:

.. code-block:: text

    http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

If IAM roles are not used you need to specify them either in a pillar or
in the minion's config file:

.. code-block:: yaml

    asg.keyid: GKTADJGHEIQSXMKKRBJ08H
    asg.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify key, keyid and region via a profile, either
as a passed in dict, or as a string to pull from pillars or minion config:

.. code-block:: yaml

    myprofile:
        keyid: GKTADJGHEIQSXMKKRBJ08H
        key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
        region: us-east-1

.. code-block:: yaml

    Ensure myasg exists:
      boto_asg.present:
        - name: myasg
        - launch_config_name: mylc
        - availability_zones:
          - us-east-1a
          - us-east-1b
        - min_size: 1
        - max_size: 1
        - desired_capacity: 1
        - load_balancers:
          - myelb
        - suspended_processes:
            - AddToLoadBalancer
            - AlarmNotification
        - scaling_policies
            - adjustment_type: ChangeInCapacity
            - as_name: api-production-iad
            - cooldown: 1800
            - min_adjustment_step: None
            - name: ScaleDown
            - scaling_adjustment: -1
        - region: us-east-1
        - keyid: GKTADJGHEIQSXMKKRBJ08H
        - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    # Using a profile from pillars.
    Ensure myasg exists:
      boto_asg.present:
        - name: myasg
        - launch_config_name: mylc
        - availability_zones:
          - us-east-1a
          - us-east-1b
        - min_size: 1
        - max_size: 1
        - desired_capacity: 1
        - load_balancers:
          - myelb
        - profile: myprofile

    # Passing in a profile.
    Ensure myasg exists:
      boto_asg.present:
        - name: myasg
        - launch_config_name: mylc
        - availability_zones:
          - us-east-1a
          - us-east-1b
        - min_size: 1
        - max_size: 1
        - desired_capacity: 1
        - load_balancers:
          - myelb
        - profile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

    # Deleting an autoscale group with running instances.
    Ensure myasg is deleted:
      boto_asg.absent:
        - name: myasg
        # If instances exist, we must force the deletion of the asg.
        - force: True

It's possible to specify cloudwatch alarms that will be setup along with the
ASG. Note the alarm name will be the name attribute defined, plus the ASG
resource name.

.. code-block:: yaml

    Ensure myasg exists:
      boto_asg.present:
        - name: myasg
        - launch_config_name: mylc
        - availability_zones:
          - us-east-1a
          - us-east-1b
        - min_size: 1
        - max_size: 1
        - desired_capacity: 1
        - load_balancers:
          - myelb
        - profile: myprofile
        - alarms:
            CPU:
              name: 'ASG CPU **MANAGED BY SALT**'
              attributes:
                metric: CPUUtilization
                namespace: AWS/EC2
                statistic: Average
                comparison: '>='
                threshold: 65.0
                period: 60
                evaluation_periods: 30
                unit: null
                description: 'ASG CPU'
                alarm_actions: [ 'arn:aws:sns:us-east-1:12345:myalarm' ]
                insufficient_data_actions: []
                ok_actions: [ 'arn:aws:sns:us-east-1:12345:myalarm' ]

You can also use alarms from pillars, and override values from the pillar
alarms by setting overrides on the resource. Note that 'boto_asg_alarms'
will be used as a default value for all resources, if defined and can be
used to ensure alarms are always set for an ASG resource.

Setting the alarms in a pillar:

.. code-block:: yaml

    my_asg_alarm:
      CPU:
        name: 'ASG CPU **MANAGED BY SALT**'
        attributes:
          metric: CPUUtilization
          namespace: AWS/EC2
          statistic: Average
          comparison: '>='
          threshold: 65.0
          period: 60
          evaluation_periods: 30
          unit: null
          description: 'ASG CPU'
          alarm_actions: [ 'arn:aws:sns:us-east-1:12345:myalarm' ]
          insufficient_data_actions: []
          ok_actions: [ 'arn:aws:sns:us-east-1:12345:myalarm' ]

Overriding the alarm values on the resource:

.. code-block:: yaml

    Ensure myasg exists:
      boto_asg.present:
        - name: myasg
        - launch_config_name: mylc
        - availability_zones:
          - us-east-1a
          - us-east-1b
        - min_size: 1
        - max_size: 1
        - desired_capacity: 1
        - load_balancers:
          - myelb
        - profile: myprofile
        - alarms_from_pillar: my_asg_alarm
        # override CPU:attributes:threshold
        - alarms:
            CPU:
              attributes:
                threshold: 50.0
"""

import copy
import hashlib
import logging

import salt.utils.dictupdate as dictupdate
import salt.utils.stringutils
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto is available.
    """
    if "boto_asg.exists" in __salt__:
        return "boto_asg"
    return (False, "boto_asg module could not be loaded")


def present(
    name,
    launch_config_name,
    availability_zones,
    min_size,
    max_size,
    launch_config=None,
    desired_capacity=None,
    load_balancers=None,
    default_cooldown=None,
    health_check_type=None,
    health_check_period=None,
    placement_group=None,
    vpc_zone_identifier=None,
    subnet_names=None,
    tags=None,
    termination_policies=None,
    termination_policies_from_pillar="boto_asg_termination_policies",
    suspended_processes=None,
    scaling_policies=None,
    scaling_policies_from_pillar="boto_asg_scaling_policies",
    scheduled_actions=None,
    scheduled_actions_from_pillar="boto_asg_scheduled_actions",
    alarms=None,
    alarms_from_pillar="boto_asg_alarms",
    region=None,
    key=None,
    keyid=None,
    profile=None,
    notification_arn=None,
    notification_arn_from_pillar="boto_asg_notification_arn",
    notification_types=None,
    notification_types_from_pillar="boto_asg_notification_types",
):
    """
    Ensure the autoscale group exists.

    name
        Name of the autoscale group.

    launch_config_name
        Name of the launch config to use for the group.  Or, if
        ``launch_config`` is specified, this will be the launch config
        name's prefix.  (see below)

    launch_config
        A dictionary of launch config attributes.  If specified, a
        launch config will be used or created, matching this set
        of attributes, and the autoscale group will be set to use
        that launch config.  The launch config name will be the
        ``launch_config_name`` followed by a hyphen followed by a hash
        of the ``launch_config`` dict contents.
        Example:

        .. code-block:: yaml

            my_asg:
              boto_asg.present:
              - launch_config:
                - ebs_optimized: false
                - instance_profile_name: my_iam_profile
                - kernel_id: ''
                - ramdisk_id: ''
                - key_name: my_ssh_key
                - image_name: aws2015091-hvm
                - instance_type: c3.xlarge
                - instance_monitoring: false
                - security_groups:
                  - my_sec_group_01
                  - my_sec_group_02

    availability_zones
        List of availability zones for the group.

    min_size
        Minimum size of the group.

    max_size
        Maximum size of the group.

    desired_capacity
        The desired capacity of the group.

    load_balancers
        List of load balancers for the group. Once set this can not be
        updated (Amazon restriction).

    default_cooldown
        Number of seconds after a Scaling Activity completes before any further
        scaling activities can start.

    health_check_type
        The service you want the health status from, Amazon EC2 or Elastic Load
        Balancer (EC2 or ELB).

    health_check_period
        Length of time in seconds after a new EC2 instance comes into service
        that Auto Scaling starts checking its health.

    placement_group
        Physical location of your cluster placement group created in Amazon
        EC2. Once set this can not be updated (Amazon restriction).

    vpc_zone_identifier
        A list of the subnet identifiers of the Virtual Private Cloud.

    subnet_names
        For VPC, a list of subnet names (NOT subnet IDs) to deploy into.
        Exclusive with vpc_zone_identifier.

    tags
        A list of tags. Example:

        .. code-block:: yaml

            - key: 'key'
              value: 'value'
              propagate_at_launch: true

    termination_policies
        A list of termination policies. Valid values are:

        * ``OldestInstance``
        * ``NewestInstance``
        * ``OldestLaunchConfiguration``
        * ``ClosestToNextInstanceHour``
        * ``Default``

        If no value is specified, the ``Default`` value is used.

    termination_policies_from_pillar:
        name of pillar dict that contains termination policy settings.   Termination policies
        defined for this specific state will override those from pillar.

    suspended_processes
        List of processes to be suspended. see
        http://docs.aws.amazon.com/AutoScaling/latest/DeveloperGuide/US_SuspendResume.html

    scaling_policies
        List of scaling policies.  Each policy is a dict of key-values described by
        https://boto.readthedocs.io/en/latest/ref/autoscale.html#boto.ec2.autoscale.policy.ScalingPolicy

    scaling_policies_from_pillar:
        name of pillar dict that contains scaling policy settings.   Scaling policies defined for
        this specific state will override those from pillar.

    scheduled_actions:
        a dictionary of scheduled actions. Each key is the name of scheduled action and each value
        is dictionary of options. For example:

        .. code-block:: yaml

            - scheduled_actions:
                scale_up_at_10:
                    desired_capacity: 4
                    min_size: 3
                    max_size: 5
                    recurrence: "0 9 * * 1-5"
                scale_down_at_7:
                    desired_capacity: 1
                    min_size: 1
                    max_size: 1
                    recurrence: "0 19 * * 1-5"

    scheduled_actions_from_pillar:
        name of pillar dict that contains scheduled_actions settings. Scheduled actions
        for this specific state will override those from pillar.

    alarms:
        a dictionary of name->boto_cloudwatch_alarm sections to be associated with this ASG.
        All attributes should be specified except for dimension which will be
        automatically set to this ASG.

        See the :mod:`salt.states.boto_cloudwatch_alarm` state for information
        about these attributes.

        If any alarm actions include  ":self:" this will be replaced with the asg name.
        For example, alarm_actions reading "['scaling_policy:self:ScaleUp']" will
        map to the arn for this asg's scaling policy named "ScaleUp".
        In addition, any alarms that have only scaling_policy as actions will be ignored if
        min_size is equal to max_size for this ASG.

    alarms_from_pillar:
        name of pillar dict that contains alarm settings.   Alarms defined for this specific
        state will override those from pillar.

    region
        The region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.

    notification_arn
        The AWS arn that notifications will be sent to

    notification_arn_from_pillar
        name of the pillar dict that contains ``notifcation_arn`` settings.  A
        ``notification_arn`` defined for this specific state will override the
        one from pillar.

    notification_types
        A list of event names that will trigger a notification.  The list of valid
        notification types is:

        * ``autoscaling:EC2_INSTANCE_LAUNCH``
        * ``autoscaling:EC2_INSTANCE_LAUNCH_ERROR``
        * ``autoscaling:EC2_INSTANCE_TERMINATE``
        * ``autoscaling:EC2_INSTANCE_TERMINATE_ERROR``
        * ``autoscaling:TEST_NOTIFICATION``

    notification_types_from_pillar
        name of the pillar dict that contains ``notifcation_types`` settings.
        ``notification_types`` defined for this specific state will override those
        from the pillar.
    """
    if vpc_zone_identifier and subnet_names:
        raise SaltInvocationError(
            "vpc_zone_identifier and subnet_names are mutually exclusive options."
        )
    ret = {"name": name, "result": True, "comment": "", "changes": {}}
    if subnet_names:
        vpc_zone_identifier = []
        for i in subnet_names:
            r = __salt__["boto_vpc.get_resource_id"](
                "subnet", name=i, region=region, key=key, keyid=keyid, profile=profile
            )
            if "error" in r:
                ret["comment"] = "Error looking up subnet ids: {}".format(r["error"])
                ret["result"] = False
                return ret
            if "id" not in r:
                ret["comment"] = f"Subnet {i} does not exist."
                ret["result"] = False
                return ret
            vpc_zone_identifier.append(r["id"])
    if vpc_zone_identifier:
        vpc_id = __salt__["boto_vpc.get_subnet_association"](
            vpc_zone_identifier, region, key, keyid, profile
        )
        vpc_id = vpc_id.get("vpc_id")
        log.debug("Auto Scaling Group %s is associated with VPC ID %s", name, vpc_id)
    else:
        vpc_id = None
        log.debug("Auto Scaling Group %s has no VPC Association", name)
    # if launch_config is defined, manage the launch config first.
    # hash the launch_config dict to create a unique name suffix and then
    # ensure it is present
    if launch_config:
        launch_config_bytes = salt.utils.stringutils.to_bytes(str(launch_config))
        launch_config_name = (
            launch_config_name + "-" + hashlib.md5(launch_config_bytes).hexdigest()
        )
        args = {
            "name": launch_config_name,
            "region": region,
            "key": key,
            "keyid": keyid,
            "profile": profile,
        }

        for index, item in enumerate(launch_config):
            if "image_name" in item:
                image_name = item["image_name"]
                iargs = {
                    "ami_name": image_name,
                    "region": region,
                    "key": key,
                    "keyid": keyid,
                    "profile": profile,
                }
                image_ids = __salt__["boto_ec2.find_images"](**iargs)
                if image_ids:  # find_images() returns False on failure
                    launch_config[index]["image_id"] = image_ids[0]
                else:
                    log.warning(
                        "Couldn't find AMI named `%s`, passing literally.", image_name
                    )
                    launch_config[index]["image_id"] = image_name
                del launch_config[index]["image_name"]
                break

        if vpc_id:
            log.debug("Auto Scaling Group {0} is a associated with a vpc")
            # locate the security groups attribute of a launch config
            sg_index = None
            for index, item in enumerate(launch_config):
                if "security_groups" in item:
                    sg_index = index
                    break
            # if security groups exist within launch_config then convert
            # to group ids
            if sg_index is not None:
                log.debug("security group associations found in launch config")
                _group_ids = __salt__["boto_secgroup.convert_to_group_ids"](
                    launch_config[sg_index]["security_groups"],
                    vpc_id=vpc_id,
                    region=region,
                    key=key,
                    keyid=keyid,
                    profile=profile,
                )
                launch_config[sg_index]["security_groups"] = _group_ids

        for d in launch_config:
            args.update(d)
        if not __opts__["test"]:
            lc_ret = __states__["boto_lc.present"](**args)
            if lc_ret["result"] is True and lc_ret["changes"]:
                if "launch_config" not in ret["changes"]:
                    ret["changes"]["launch_config"] = {}
                ret["changes"]["launch_config"] = lc_ret["changes"]

    asg = __salt__["boto_asg.get_config"](name, region, key, keyid, profile)
    termination_policies = _determine_termination_policies(
        termination_policies, termination_policies_from_pillar
    )
    scaling_policies = _determine_scaling_policies(
        scaling_policies, scaling_policies_from_pillar
    )
    scheduled_actions = _determine_scheduled_actions(
        scheduled_actions, scheduled_actions_from_pillar
    )
    if asg is None:
        ret["result"] = False
        ret["comment"] = "Failed to check autoscale group existence."
    elif not asg:
        if __opts__["test"]:
            msg = "Autoscale group set to be created."
            ret["comment"] = msg
            ret["result"] = None
            return ret
        notification_arn, notification_types = _determine_notification_info(
            notification_arn,
            notification_arn_from_pillar,
            notification_types,
            notification_types_from_pillar,
        )
        created = __salt__["boto_asg.create"](
            name,
            launch_config_name,
            availability_zones,
            min_size,
            max_size,
            desired_capacity,
            load_balancers,
            default_cooldown,
            health_check_type,
            health_check_period,
            placement_group,
            vpc_zone_identifier,
            tags,
            termination_policies,
            suspended_processes,
            scaling_policies,
            scheduled_actions,
            region,
            notification_arn,
            notification_types,
            key,
            keyid,
            profile,
        )
        if created:
            ret["changes"]["old"] = None
            asg = __salt__["boto_asg.get_config"](name, region, key, keyid, profile)
            ret["changes"]["new"] = asg
        else:
            ret["result"] = False
            ret["comment"] = "Failed to create autoscale group"
    else:
        need_update = False
        # If any of these attributes can't be modified after creation
        # time, we should remove them from the dict.
        if scaling_policies:
            for policy in scaling_policies:
                if "min_adjustment_step" not in policy:
                    policy["min_adjustment_step"] = None
        if scheduled_actions:
            for s_name, action in scheduled_actions.items():
                if "end_time" not in action:
                    action["end_time"] = None
        config = {
            "launch_config_name": launch_config_name,
            "availability_zones": availability_zones,
            "min_size": min_size,
            "max_size": max_size,
            "desired_capacity": desired_capacity,
            "default_cooldown": default_cooldown,
            "health_check_type": health_check_type,
            "health_check_period": health_check_period,
            "vpc_zone_identifier": vpc_zone_identifier,
            "tags": tags,
            "termination_policies": termination_policies,
            "suspended_processes": suspended_processes,
            "scaling_policies": scaling_policies,
            "scheduled_actions": scheduled_actions,
        }
        # ensure that we reset termination_policies to default if none are specified
        if not termination_policies:
            config["termination_policies"] = ["Default"]
        if suspended_processes is None:
            config["suspended_processes"] = []
        # ensure that we delete scaling_policies if none are specified
        if scaling_policies is None:
            config["scaling_policies"] = []
        # ensure that we delete scheduled_actions if none are specified
        if scheduled_actions is None:
            config["scheduled_actions"] = {}
        # allow defaults on start_time
        for s_name, action in scheduled_actions.items():
            if "start_time" not in action:
                asg_action = asg["scheduled_actions"].get(s_name, {})
                if "start_time" in asg_action:
                    del asg_action["start_time"]
        proposed = {}
        # note: do not loop using "key, value" - this can modify the value of
        # the aws access key
        for asg_property, value in config.items():
            # Only modify values being specified; introspection is difficult
            # otherwise since it's hard to track default values, which will
            # always be returned from AWS.
            if value is None:
                continue
            value = __utils__["boto3.ordered"](value)
            if asg_property in asg:
                _value = __utils__["boto3.ordered"](asg[asg_property])
                if not value == _value:
                    log.debug("%s asg_property differs from %s", value, _value)
                    proposed.setdefault("old", {}).update({asg_property: _value})
                    proposed.setdefault("new", {}).update({asg_property: value})
                    need_update = True
        if need_update:
            if __opts__["test"]:
                msg = "Autoscale group set to be updated."
                ret["comment"] = msg
                ret["result"] = None
                ret["changes"] = proposed
                return ret
            # add in alarms
            notification_arn, notification_types = _determine_notification_info(
                notification_arn,
                notification_arn_from_pillar,
                notification_types,
                notification_types_from_pillar,
            )
            updated, msg = __salt__["boto_asg.update"](
                name,
                launch_config_name,
                availability_zones,
                min_size,
                max_size,
                desired_capacity=desired_capacity,
                load_balancers=load_balancers,
                default_cooldown=default_cooldown,
                health_check_type=health_check_type,
                health_check_period=health_check_period,
                placement_group=placement_group,
                vpc_zone_identifier=vpc_zone_identifier,
                tags=tags,
                termination_policies=termination_policies,
                suspended_processes=suspended_processes,
                scaling_policies=scaling_policies,
                scheduled_actions=scheduled_actions,
                region=region,
                notification_arn=notification_arn,
                notification_types=notification_types,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if asg["launch_config_name"] != launch_config_name:
                # delete the old launch_config_name
                deleted = __salt__["boto_asg.delete_launch_configuration"](
                    asg["launch_config_name"],
                    region=region,
                    key=key,
                    keyid=keyid,
                    profile=profile,
                )
                if deleted:
                    if "launch_config" not in ret["changes"]:
                        ret["changes"]["launch_config"] = {}
                    ret["changes"]["launch_config"]["deleted"] = asg[
                        "launch_config_name"
                    ]
            if updated:
                ret["changes"]["old"] = asg
                asg = __salt__["boto_asg.get_config"](name, region, key, keyid, profile)
                ret["changes"]["new"] = asg
                ret["comment"] = "Updated autoscale group."
            else:
                ret["result"] = False
                ret["comment"] = msg
        else:
            ret["comment"] = "Autoscale group present."
    # add in alarms
    _ret = _alarms_present(
        name,
        min_size == max_size,
        alarms,
        alarms_from_pillar,
        region,
        key,
        keyid,
        profile,
    )
    ret["changes"] = dictupdate.update(ret["changes"], _ret["changes"])
    ret["comment"] = " ".join([ret["comment"], _ret["comment"]])
    if not _ret["result"]:
        ret["result"] = _ret["result"]
    return ret


def _determine_termination_policies(
    termination_policies, termination_policies_from_pillar
):
    """
    helper method for present.  ensure that termination_policies are set
    """
    pillar_termination_policies = copy.deepcopy(
        __salt__["config.option"](termination_policies_from_pillar, [])
    )
    if not termination_policies and len(pillar_termination_policies) > 0:
        termination_policies = pillar_termination_policies
    return termination_policies


def _determine_scaling_policies(scaling_policies, scaling_policies_from_pillar):
    """
    helper method for present.  ensure that scaling_policies are set
    """
    pillar_scaling_policies = copy.deepcopy(
        __salt__["config.option"](scaling_policies_from_pillar, {})
    )
    if not scaling_policies and len(pillar_scaling_policies) > 0:
        scaling_policies = pillar_scaling_policies
    return scaling_policies


def _determine_scheduled_actions(scheduled_actions, scheduled_actions_from_pillar):
    """
    helper method for present,  ensure scheduled actions are setup
    """
    tmp = copy.deepcopy(__salt__["config.option"](scheduled_actions_from_pillar, {}))
    # merge with data from state
    if scheduled_actions:
        tmp = dictupdate.update(tmp, scheduled_actions)
    return tmp


def _determine_notification_info(
    notification_arn,
    notification_arn_from_pillar,
    notification_types,
    notification_types_from_pillar,
):
    """
    helper method for present.  ensure that notification_configs are set
    """
    pillar_arn_list = copy.deepcopy(
        __salt__["config.option"](notification_arn_from_pillar, {})
    )
    pillar_arn = None
    if len(pillar_arn_list) > 0:
        pillar_arn = pillar_arn_list[0]
    pillar_notification_types = copy.deepcopy(
        __salt__["config.option"](notification_types_from_pillar, {})
    )
    arn = notification_arn if notification_arn else pillar_arn
    types = notification_types if notification_types else pillar_notification_types
    return (arn, types)


def _alarms_present(
    name,
    min_size_equals_max_size,
    alarms,
    alarms_from_pillar,
    region,
    key,
    keyid,
    profile,
):
    """
    helper method for present.  ensure that cloudwatch_alarms are set
    """
    # load data from alarms_from_pillar
    tmp = copy.deepcopy(__salt__["config.option"](alarms_from_pillar, {}))
    # merge with data from alarms
    if alarms:
        tmp = dictupdate.update(tmp, alarms)
    # set alarms, using boto_cloudwatch_alarm.present
    merged_return_value = {"name": name, "result": True, "comment": "", "changes": {}}
    for _, info in tmp.items():
        # add asg to name and description
        info["name"] = name + " " + info["name"]
        info["attributes"]["description"] = (
            name + " " + info["attributes"]["description"]
        )
        # add dimension attribute
        if "dimensions" not in info["attributes"]:
            info["attributes"]["dimensions"] = {"AutoScalingGroupName": [name]}
        scaling_policy_actions_only = True
        # replace ":self:" with our name
        for action_type in ["alarm_actions", "insufficient_data_actions", "ok_actions"]:
            if action_type in info["attributes"]:
                new_actions = []
                for action in info["attributes"][action_type]:
                    if "scaling_policy" not in action:
                        scaling_policy_actions_only = False
                    if ":self:" in action:
                        action = action.replace(":self:", f":{name}:")
                    new_actions.append(action)
                info["attributes"][action_type] = new_actions
        # skip alarms that only have actions for scaling policy, if min_size == max_size for this ASG
        if scaling_policy_actions_only and min_size_equals_max_size:
            continue
        # set alarm
        kwargs = {
            "name": info["name"],
            "attributes": info["attributes"],
            "region": region,
            "key": key,
            "keyid": keyid,
            "profile": profile,
        }
        results = __states__["boto_cloudwatch_alarm.present"](**kwargs)
        if not results["result"]:
            merged_return_value["result"] = False
        if results.get("changes", {}) != {}:
            merged_return_value["changes"][info["name"]] = results["changes"]
        if "comment" in results:
            merged_return_value["comment"] += results["comment"]
    return merged_return_value


def absent(
    name, force=False, region=None, key=None, keyid=None, profile=None, remove_lc=False
):
    """
    Ensure the named autoscale group is deleted.

    name
        Name of the autoscale group.

    force
        Force deletion of autoscale group.

    remove_lc
        Delete the launch config as well.

    region
        The region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}
    asg = __salt__["boto_asg.get_config"](name, region, key, keyid, profile)
    if asg is None:
        ret["result"] = False
        ret["comment"] = "Failed to check autoscale group existence."
    elif asg:
        if __opts__["test"]:
            ret["comment"] = "Autoscale group set to be deleted."
            ret["result"] = None
            if remove_lc:
                msg = "Launch configuration {} is set to be deleted.".format(
                    asg["launch_config_name"]
                )
                ret["comment"] = " ".join([ret["comment"], msg])
            return ret
        deleted = __salt__["boto_asg.delete"](name, force, region, key, keyid, profile)
        if deleted:
            if remove_lc:
                lc_deleted = __salt__["boto_asg.delete_launch_configuration"](
                    asg["launch_config_name"], region, key, keyid, profile
                )
                if lc_deleted:
                    if "launch_config" not in ret["changes"]:
                        ret["changes"]["launch_config"] = {}
                    ret["changes"]["launch_config"]["deleted"] = asg[
                        "launch_config_name"
                    ]
                else:
                    ret["result"] = False
                    ret["comment"] = " ".join(
                        [ret["comment"], "Failed to delete launch configuration."]
                    )
            ret["changes"]["old"] = asg
            ret["changes"]["new"] = None
            ret["comment"] = "Deleted autoscale group."
        else:
            ret["result"] = False
            ret["comment"] = "Failed to delete autoscale group."
    else:
        ret["comment"] = "Autoscale group does not exist."
    return ret



"""
checkrestart functionality for Debian and Red Hat Based systems

Identifies services (processes) that are linked against deleted files (for example after downloading an updated
binary of a shared library).

Based on checkrestart script from debian-goodies (written  by Matt Zimmerman for the Debian GNU/Linux distribution,
https://packages.debian.org/debian-goodies) and psdel by Sam Morris.

:codeauthor: Jiri Kotlin <jiri.kotlin@ultimum.io>
"""

import os
import re
import subprocess
import sys
import time

import salt.exceptions
import salt.utils.args
import salt.utils.files
import salt.utils.path

NILRT_FAMILY_NAME = "NILinuxRT"

HAS_PSUTIL = False
try:
    import psutil

    HAS_PSUTIL = True
except ImportError:
    pass

LIST_DIRS = [
    # We don't care about log files
    "^/var/log/",
    "^/var/local/log/",
    # Or about files under temporary locations
    "^/var/run/",
    "^/var/local/run/",
    # Or about files under /tmp
    "^/tmp/",
    # Or about files under /dev/shm
    "^/dev/shm/",
    # Or about files under /run
    "^/run/",
    # Or about files under /drm
    "^/drm",
    # Or about files under /var/tmp and /var/local/tmp
    "^/var/tmp/",
    "^/var/local/tmp/",
    # Or /dev/zero
    "^/dev/zero",
    # Or /dev/pts (used by gpm)
    "^/dev/pts/",
    # Or /usr/lib/locale
    "^/usr/lib/locale/",
    # Skip files from the user's home directories
    # many processes hold temporafy files there
    "^/home/",
    # Skip automatically generated files
    "^.*icon-theme.cache",
    # Skip font files
    "^/var/cache/fontconfig/",
    # Skip Nagios Spool
    "^/var/lib/nagios3/spool/",
    # Skip nagios spool files
    "^/var/lib/nagios3/spool/checkresults/",
    # Skip Postgresql files
    "^/var/lib/postgresql/",
    # Skip VDR lib files
    "^/var/lib/vdr/",
    # Skip Aio files found in MySQL servers
    "^/[aio]",
    # ignore files under /SYSV
    "^/SYSV",
]


def __virtual__():
    """
    Only run this module if the psutil python module is installed (package python-psutil).
    """
    if HAS_PSUTIL:
        return HAS_PSUTIL
    else:
        return (False, "Missing dependency: psutil")


def _valid_deleted_file(path):
    """
    Filters file path against unwanted directories and decides whether file is marked as deleted.

    Returns:
        True if file is desired deleted file, else False.

    Args:
        path: A string - path to file.
    """
    ret = False
    if path.endswith(" (deleted)"):
        ret = True
    if re.compile(r"\(path inode=[0-9]+\)$").search(path):
        ret = True

    regex = re.compile("|".join(LIST_DIRS))
    if regex.match(path):
        ret = False
    return ret


def _deleted_files():
    """
    Iterates over /proc/PID/maps and /proc/PID/fd links and returns list of desired deleted files.

    Returns:
        List of deleted files to analyze, False on failure.

    """
    deleted_files = []

    for proc in psutil.process_iter():  # pylint: disable=too-many-nested-blocks
        try:
            pinfo = proc.as_dict(attrs=["pid", "name"])
            try:
                with salt.utils.files.fopen(
                    "/proc/{}/maps".format(pinfo["pid"])
                ) as maps:  # pylint: disable=resource-leakage
                    dirpath = "/proc/" + str(pinfo["pid"]) + "/fd/"
                    listdir = os.listdir(dirpath)
                    maplines = maps.readlines()
            except OSError:
                yield False

            # /proc/PID/maps
            mapline = re.compile(
                r"^[\da-f]+-[\da-f]+ [r-][w-][x-][sp-] "
                r"[\da-f]+ [\da-f]{2}:[\da-f]{2} (\d+) *(.+)( \(deleted\))?\n$"
            )

            for line in maplines:
                line = salt.utils.stringutils.to_unicode(line)
                matched = mapline.match(line)
                if not matched:
                    continue
                path = matched.group(2)
                if not path:
                    continue
                valid = _valid_deleted_file(path)
                if not valid:
                    continue
                val = (pinfo["name"], pinfo["pid"], path[0:-10])
                if val not in deleted_files:
                    deleted_files.append(val)
                    yield val

            # /proc/PID/fd
            try:
                for link in listdir:
                    path = dirpath + link
                    readlink = os.readlink(path)
                    filenames = []

                    if os.path.isfile(readlink):
                        filenames.append(readlink)
                    elif os.path.isdir(readlink) and readlink != "/":
                        for root, dummy_dirs, files in salt.utils.path.os_walk(
                            readlink, followlinks=True
                        ):
                            for name in files:
                                filenames.append(os.path.join(root, name))

                    for filename in filenames:
                        valid = _valid_deleted_file(filename)
                        if not valid:
                            continue
                        val = (pinfo["name"], pinfo["pid"], filename)
                        if val not in deleted_files:
                            deleted_files.append(val)
                            yield val
            except OSError:
                pass

        except psutil.NoSuchProcess:
            pass


def _format_output(
    kernel_restart,
    packages,
    verbose,
    restartable,
    nonrestartable,
    restartservicecommands,
    restartinitcommands,
):
    """
    Formats the output of the restartcheck module.

    Returns:
        String - formatted output.

    Args:
        kernel_restart: indicates that newer kernel is instaled.
        packages: list of packages that should be restarted.
        verbose: enables extensive output.
        restartable: list of restartable packages.
        nonrestartable: list of non-restartable packages.
        restartservicecommands: list of commands to restart services.
        restartinitcommands: list of commands to restart init.d scripts.

    """
    if not verbose:
        packages = restartable + nonrestartable
        if kernel_restart:
            packages.append("System restart required.")
        return packages
    else:
        ret = ""
        if kernel_restart:
            ret = "System restart required.\n\n"

        if packages:
            ret += "Found {} processes using old versions of upgraded files.\n".format(
                len(packages)
            )
            ret += "These are the packages:\n"

        if restartable:
            ret += (
                "Of these, {} seem to contain systemd service definitions or init"
                " scripts which can be used to restart them:\n".format(len(restartable))
            )
            for package in restartable:
                ret += package + ":\n"
                for program in packages[package]["processes"]:
                    ret += program + "\n"

            if restartservicecommands:
                ret += "\n\nThese are the systemd services:\n"
                ret += "\n".join(restartservicecommands)

            if restartinitcommands:
                ret += "\n\nThese are the initd scripts:\n"
                ret += "\n".join(restartinitcommands)

        if nonrestartable:
            ret += (
                "\n\nThese processes {} do not seem to have an associated init script "
                "to restart them:\n".format(len(nonrestartable))
            )
            for package in nonrestartable:
                ret += package + ":\n"
                for program in packages[package]["processes"]:
                    ret += program + "\n"
    return ret


def _kernel_versions_debian():
    """
    Last installed kernel name, for Debian based systems.

    Returns:
            List with possible names of last installed kernel
            as they are probably interpreted in output of `uname -a` command.
    """
    kernel_get_selections = __salt__["cmd.run"]("dpkg --get-selections linux-image-*")
    kernels = []
    kernel_versions = []
    for line in kernel_get_selections.splitlines():
        kernels.append(line)

    try:
        kernel = kernels[-2]
    except IndexError:
        kernel = kernels[0]

    kernel = kernel.rstrip("\t\tinstall")

    kernel_get_version = __salt__["cmd.run"]("apt-cache policy " + kernel)

    for line in kernel_get_version.splitlines():
        if line.startswith("  Installed: "):
            kernel_v = line.strip("  Installed: ")
            kernel_versions.append(kernel_v)
            break

    if __grains__["os"] == "Ubuntu":
        kernel_v = kernel_versions[0].rsplit(".", 1)
        kernel_ubuntu_generic = kernel_v[0] + "-generic #" + kernel_v[1]
        kernel_ubuntu_lowlatency = kernel_v[0] + "-lowlatency #" + kernel_v[1]
        kernel_versions.extend([kernel_ubuntu_generic, kernel_ubuntu_lowlatency])

    return kernel_versions


def _kernel_versions_redhat():
    """
    Name of the last installed kernel, for Red Hat based systems.

    Returns:
            List with name of last installed kernel as it is interpreted in output of `uname -a` command.
    """
    kernel_get_last = __salt__["cmd.run"]("rpm -q --last kernel")
    kernels = []
    kernel_versions = []
    for line in kernel_get_last.splitlines():
        if "kernel-" in line:
            kernels.append(line)

    kernel = kernels[0].split(" ", 1)[0]
    kernel = kernel.strip("kernel-")
    kernel_versions.append(kernel)

    return kernel_versions


def _kernel_versions_nilrt():
    """
    Last installed kernel name, for Debian based systems.

    Returns:
            List with possible names of last installed kernel
            as they are probably interpreted in output of `uname -a` command.
    """
    kver = None

    def _get_kver_from_bin(kbin):
        """
        Get kernel version from a binary image or None if detection fails
        """
        kvregex = r"[0-9]+\.[0-9]+\.[0-9]+-rt\S+"
        kernel_strings = __salt__["cmd.run"](f"strings {kbin}")
        re_result = re.search(kvregex, kernel_strings)
        return None if re_result is None else re_result.group(0)

    if __grains__.get("lsb_distrib_id") == "nilrt":
        if "arm" in __grains__.get("cpuarch"):
            # the kernel is inside a uboot created itb (FIT) image alongside the
            # device tree, ramdisk and a bootscript. There is no package management
            # or any other kind of versioning info, so we need to extract the itb.
            itb_path = "/boot/linux_runmode.itb"
            compressed_kernel = "/var/volatile/tmp/uImage.gz"
            uncompressed_kernel = "/var/volatile/tmp/uImage"
            __salt__["cmd.run"](
                "dumpimage -i {} -T flat_dt -p0 kernel -o {}".format(
                    itb_path, compressed_kernel
                )
            )
            __salt__["cmd.run"](f"gunzip -f {compressed_kernel}")
            kver = _get_kver_from_bin(uncompressed_kernel)
        else:
            # the kernel bzImage is copied to rootfs without package management or
            # other versioning info.
            kver = _get_kver_from_bin("/boot/runmode/bzImage")
    else:
        # kernels in newer NILRT's are installed via package management and
        # have the version appended to the kernel image filename
        if "arm" in __grains__.get("cpuarch"):
            kver = os.path.basename(os.readlink("/boot/uImage")).strip("uImage-")
        else:
            kver = os.path.basename(os.readlink("/boot/bzImage")).strip("bzImage-")

    return [] if kver is None else [kver]


def _check_timeout(start_time, timeout):
    """
    Name of the last installed kernel, for Red Hat based systems.

    Returns:
            List with name of last installed kernel as it is interpreted in output of `uname -a` command.
    """
    timeout_milisec = timeout * 60000
    if timeout_milisec < (int(round(time.time() * 1000)) - start_time):
        raise salt.exceptions.TimeoutError("Timeout expired.")


def _file_changed_nilrt(full_filepath):
    """
    Detect whether a file changed in an NILinuxRT system using md5sum and timestamp
    files from a state directory.

    Returns:
             - True if either md5sum/timestamp state files do not exist, or
               the file at ``full_filepath`` was touched or modified.
             - False otherwise.
    """
    rs_state_dir = "/var/lib/salt/restartcheck_state"
    base_filename = os.path.basename(full_filepath)
    timestamp_file = os.path.join(rs_state_dir, f"{base_filename}.timestamp")
    md5sum_file = os.path.join(rs_state_dir, f"{base_filename}.md5sum")

    if not os.path.exists(timestamp_file) or not os.path.exists(md5sum_file):
        return True

    prev_timestamp = __salt__["file.read"](timestamp_file).rstrip()
    # Need timestamp in seconds so floor it using int()
    cur_timestamp = str(int(os.path.getmtime(full_filepath)))

    if prev_timestamp != cur_timestamp:
        return True

    return bool(
        __salt__["cmd.retcode"](f"md5sum -cs {md5sum_file}", output_loglevel="quiet")
    )


def _kernel_modules_changed_nilrt(kernelversion):
    """
    Once a NILRT kernel module is inserted, it can't be rmmod so systems need
    rebooting (some modules explicitly ask for reboots even on first install),
    hence this functionality of determining if the module state got modified by
    testing if depmod was run.

    Returns:
             - True if modules.dep was modified/touched, False otherwise.
    """
    if kernelversion is not None:
        return _file_changed_nilrt(f"/lib/modules/{kernelversion}/modules.dep")
    return False


def _sysapi_changed_nilrt():
    """
    Besides the normal Linux kernel driver interfaces, NILinuxRT-supported hardware features an
    extensible, plugin-based device enumeration and configuration interface named "System API".
    When an installed package is extending the API it is very hard to know all repercurssions and
    actions to be taken, so reboot making sure all drivers are reloaded, hardware reinitialized,
    daemons restarted, etc.

    Returns:
             - True if nisysapi .ini files were modified/touched.
             - False if no nisysapi .ini files exist.
    """
    nisysapi_path = "/usr/local/natinst/share/nisysapi.ini"
    if os.path.exists(nisysapi_path) and _file_changed_nilrt(nisysapi_path):
        return True

    restartcheck_state_dir = "/var/lib/salt/restartcheck_state"
    nisysapi_conf_d_path = "/usr/lib/{}/nisysapi/conf.d/experts/".format(
        "arm-linux-gnueabi"
        if "arm" in __grains__.get("cpuarch")
        else "x86_64-linux-gnu"
    )

    if os.path.exists(nisysapi_conf_d_path):
        rs_count_file = f"{restartcheck_state_dir}/sysapi.conf.d.count"
        if not os.path.exists(rs_count_file):
            return True

        with salt.utils.files.fopen(rs_count_file, "r") as fcount:
            current_nb_files = len(os.listdir(nisysapi_conf_d_path))
            rs_stored_nb_files = int(fcount.read())
            if current_nb_files != rs_stored_nb_files:
                return True

        for fexpert in os.listdir(nisysapi_conf_d_path):
            if _file_changed_nilrt(f"{nisysapi_conf_d_path}/{fexpert}"):
                return True

    return False


# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def restartcheck(ignorelist=None, blacklist=None, excludepid=None, **kwargs):
    """
    Analyzes files openeded by running processes and seeks for packages which need to be restarted.

    Args:
        ignorelist: string or list of packages to be ignored.
        blacklist: string or list of file paths to be ignored.
        excludepid: string or list of process IDs to be ignored.
        verbose: boolean, enables extensive output.
        timeout: int, timeout in minute.

    Returns:
        Dict on error: ``{ 'result': False, 'comment': '<reason>' }``.
        String with checkrestart output if some package seems to need to be restarted or
        if no packages need restarting.

    .. versionadded:: 2015.8.3

    CLI Example:

    .. code-block:: bash

        salt '*' restartcheck.restartcheck
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    start_time = int(round(time.time() * 1000))
    kernel_restart = True
    verbose = kwargs.pop("verbose", True)
    timeout = kwargs.pop("timeout", 5)
    if __grains__.get("os_family") == "Debian":
        cmd_pkg_query = ["dpkg-query", "--listfiles"]
        systemd_folder = "/lib/systemd/system/"
        systemd = "/bin/systemd"
        kernel_versions = _kernel_versions_debian()
    elif __grains__.get("os_family") == "RedHat":
        cmd_pkg_query = ["repoquery", "-l"]
        systemd_folder = "/usr/lib/systemd/system/"
        systemd = "/usr/bin/systemctl"
        kernel_versions = _kernel_versions_redhat()
    elif __grains__.get("os_family") == NILRT_FAMILY_NAME:
        cmd_pkg_query = ["opkg", "files"]
        systemd = ""
        kernel_versions = _kernel_versions_nilrt()
    else:
        return {
            "result": False,
            "comment": (
                "Only available on Debian, Red Hat and NI Linux Real-Time based"
                " systems."
            ),
        }

    # Check kernel versions
    kernel_current = __salt__["cmd.run"]("uname -a")
    for kernel in kernel_versions:
        _check_timeout(start_time, timeout)
        if kernel in kernel_current:
            if __grains__.get("os_family") == "NILinuxRT":
                # Check kernel modules and hardware API's for version changes
                # If a restartcheck=True event was previously witnessed, propagate it
                if (
                    not _kernel_modules_changed_nilrt(kernel)
                    and not _sysapi_changed_nilrt()
                    and not __salt__["system.get_reboot_required_witnessed"]()
                ):
                    kernel_restart = False
                    break
            else:
                kernel_restart = False
                break

    packages = {}
    running_services = {}
    restart_services = []

    if ignorelist:
        if not isinstance(ignorelist, list):
            ignorelist = [ignorelist]
    else:
        ignorelist = ["screen", "systemd"]

    if blacklist:
        if not isinstance(blacklist, list):
            blacklist = [blacklist]
    else:
        blacklist = []

    if excludepid:
        if not isinstance(excludepid, list):
            excludepid = [excludepid]
    else:
        excludepid = []

    for service in __salt__["service.get_running"]():
        _check_timeout(start_time, timeout)
        service_show = __salt__["service.show"](service)
        if "ExecMainPID" in service_show:
            running_services[service] = int(service_show["ExecMainPID"])

    owners_cache = {}
    for deleted_file in _deleted_files():
        if deleted_file is False:
            return {
                "result": False,
                "comment": (
                    "Could not get list of processes. (Do you have root access?)"
                ),
            }

        _check_timeout(start_time, timeout)
        name, pid, path = deleted_file[0], deleted_file[1], deleted_file[2]
        if path in blacklist or pid in excludepid:
            continue
        try:
            readlink = os.readlink(f"/proc/{pid}/exe")
        except OSError:
            excludepid.append(pid)
            continue
        try:
            packagename = owners_cache[readlink]
        except KeyError:
            packagename = __salt__["pkg.owner"](readlink)
            if not packagename:
                packagename = name
            owners_cache[readlink] = packagename
        for running_service in running_services:
            _check_timeout(start_time, timeout)
            if (
                running_service not in restart_services
                and pid == running_services[running_service]
            ):
                if packagename and packagename not in ignorelist:
                    restart_services.append(running_service)
                    name = running_service
        if packagename and packagename not in ignorelist:
            program = "\t" + str(pid) + " " + readlink + " (file: " + str(path) + ")"
            if packagename not in packages:
                packages[packagename] = {
                    "initscripts": [],
                    "systemdservice": [],
                    "processes": [program],
                    "process_name": name,
                }
            else:
                if program not in packages[packagename]["processes"]:
                    packages[packagename]["processes"].append(program)

    if not packages and not kernel_restart:
        return "No packages seem to need to be restarted."

    for package in packages:
        _check_timeout(start_time, timeout)
        cmd = cmd_pkg_query[:]
        cmd.append(package)
        paths = subprocess.Popen(cmd, stdout=subprocess.PIPE)

        while True:
            _check_timeout(start_time, timeout)
            line = salt.utils.stringutils.to_unicode(paths.stdout.readline())
            if not line:
                break
            pth = line[:-1]
            if pth.startswith("/etc/init.d/") and not pth.endswith(".sh"):
                packages[package]["initscripts"].append(pth[12:])

            if (
                os.path.exists(systemd)
                and pth.startswith(systemd_folder)
                and pth.endswith(".service")
                and pth.find(".wants") == -1
            ):
                is_oneshot = False
                try:
                    # pylint: disable=resource-leakage
                    servicefile = salt.utils.files.fopen(pth)
                    # pylint: enable=resource-leakage
                except OSError:
                    continue
                sysfold_len = len(systemd_folder)

                for line in servicefile.readlines():
                    line = salt.utils.stringutils.to_unicode(line)
                    if line.find("Type=oneshot") > 0:
                        # scripts that does a single job and then exit
                        is_oneshot = True
                    continue
                servicefile.close()

                if not is_oneshot:
                    packages[package]["systemdservice"].append(pth[sysfold_len:])

            sys.stdout.flush()
        paths.stdout.close()

    # Alternatively, find init.d script or service that match the process name
    for package in packages:
        _check_timeout(start_time, timeout)
        if (
            not packages[package]["systemdservice"]
            and not packages[package]["initscripts"]
        ):
            service = __salt__["service.available"](packages[package]["process_name"])

            if service:
                if os.path.exists("/etc/init.d/" + packages[package]["process_name"]):
                    packages[package]["initscripts"].append(
                        packages[package]["process_name"]
                    )
                else:
                    packages[package]["systemdservice"].append(
                        packages[package]["process_name"]
                    )

    restartable = []
    nonrestartable = []
    restartinitcommands = []
    restartservicecommands = []

    for package in packages:
        _check_timeout(start_time, timeout)
        if packages[package]["initscripts"]:
            restartable.append(package)
            restartinitcommands.extend(
                ["service " + s + " restart" for s in packages[package]["initscripts"]]
            )
        elif packages[package]["systemdservice"]:
            restartable.append(package)
            restartservicecommands.extend(
                ["systemctl restart " + s for s in packages[package]["systemdservice"]]
            )
        else:
            nonrestartable.append(package)
        if packages[package]["process_name"] in restart_services:
            restart_services.remove(packages[package]["process_name"])

    for restart_service in restart_services:
        _check_timeout(start_time, timeout)
        restartservicecommands.extend(["systemctl restart " + restart_service])

    ret = _format_output(
        kernel_restart,
        packages,
        verbose,
        restartable,
        nonrestartable,
        restartservicecommands,
        restartinitcommands,
    )
    return ret



import os
import re

import pytest
from saltfactories.utils import random_string

import salt.modules.win_file as win_file
import salt.modules.win_useradd
import salt.utils.user
import salt.utils.win_dacl
from salt.exceptions import CommandExecutionError
from tests.support.mock import patch

pytestmark = [pytest.mark.windows_whitelisted, pytest.mark.skip_unless_on_windows]


@pytest.fixture
def configure_loader_modules():
    return {
        win_file: {},
        salt.utils.win_dacl: {},
    }


@pytest.fixture(scope="module")
def test_user():
    user_name = random_string("test-")
    salt.modules.win_useradd.add(name=user_name, password="P@ssw0rd")
    yield user_name
    salt.modules.win_useradd.delete(name=user_name)


def test__virtual__not_windows():
    with patch("salt.utils.platform.is_windows", autospec=True, return_value=False):
        expected = (False, "Module win_file: Missing Win32 modules")
        result = win_file.__virtual__()
        assert result == expected
    with patch("salt.modules.win_file.HAS_WINDOWS_MODULES", False):
        expected = (False, "Module win_file: Missing Win32 modules")
        result = win_file.__virtual__()
        assert result == expected


def test__virtual__no_dacl():
    with patch("salt.modules.win_file.HAS_WIN_DACL", False):
        expected = (False, "Module win_file: Unable to load salt.utils.win_dacl")
        result = win_file.__virtual__()
        assert result == expected


def test__get_version_os():
    expected = ["32-bit Windows", "Windows NT"]
    result = win_file._get_version_os(0x00040004)
    assert result == expected


def test__get_version_type_application():
    expected = "Application"
    result = win_file._get_version_type(1, 0)
    assert result == expected


def test__get_version_type_driver():
    expected = "Printer Driver"
    result = win_file._get_version_type(3, 1)
    assert result == expected


def test__get_version_type_font():
    expected = "TrueType Font"
    result = win_file._get_version_type(4, 3)
    assert result == expected


def test__get_version_type_virtual_device():
    expected = "Virtual Device: 12345"
    result = win_file._get_version_type(5, 12345)
    assert result == expected


def test__get_version_exe():
    result = win_file._get_version(r"C:\Windows\notepad.exe")
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result)


def test__get_version_dll():
    result = win_file._get_version(r"C:\Windows\System32\FirewallAPI.dll")
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result)


def test__get_version_sys():
    result = win_file._get_version(r"C:\Windows\System32\drivers\netio.sys")
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result)


def test_get_pgid_error():
    with pytest.raises(CommandExecutionError):
        win_file.get_pgid("C:\\Path\\That\\Does\\Not\\Exist.txt")


def test_get_pgid():
    """
    We can't know what this value is, so we're just making sure it found
    something
    """
    result = win_file.get_pgid(os.getenv("COMSPEC"))
    assert result != ""


def test_group_to_gid():
    with patch.dict(win_file.__opts__, {}):
        result = win_file.group_to_gid("Administrators")
    expected = "S-1-5-32-544"
    assert result == expected


def test_group_to_gid_empty():
    with patch.dict(win_file.__opts__, {}):
        result = win_file.group_to_gid("")
    expected = "S-1-5-32"
    assert result == expected


def test_uid_to_user():
    result = win_file.uid_to_user("S-1-5-32-544")
    expected = "Administrators"
    assert result == expected


def test_uid_to_user_empty():
    result = win_file.uid_to_user("")
    expected = ""
    assert result == expected


def test_user_to_uid(test_user):
    result = win_file.user_to_uid(test_user)
    expected = salt.utils.win_dacl.get_sid_string(test_user)
    assert result == expected


def test_user_to_uid_none():
    result = win_file.user_to_uid(None)
    expected = salt.utils.win_dacl.get_sid_string(salt.utils.user.get_user())
    assert result == expected


def test_get_uid():
    """
    We can't know what this value is, so we're just making sure it found
    something
    """
    result = win_file.get_uid(os.getenv("WINDIR"))
    assert result != ""


def test_get_uid_error():
    with pytest.raises(CommandExecutionError):
        win_file.get_uid("C:\\fake\\path")


def test_chown(tmp_path):
    test_file = tmp_path / "test_file.txt"
    test_file.touch()
    win_file.chown(path=str(test_file), user="Administrators", pgroup="Guests")
    assert win_file.get_user(str(test_file)) == "Administrators"
    assert win_file.get_pgroup(str(test_file)) == "Guests"


def test_chpgrp(tmp_path):
    test_file = tmp_path / "test_file.txt"
    test_file.touch()
    win_file.chown(path=str(test_file), user="Administrators", pgroup="Guests")
    win_file.chpgrp(path=str(test_file), group="Administrators")
    assert win_file.get_pgroup(str(test_file)) == "Administrators"


def test_stats_mode(tmp_path):
    test_file = tmp_path / "test_file.txt"
    test_file.touch()
    results = win_file.stats(str(test_file))
    assert results["mode"] == "0666"


def test_is_link_true(tmp_path):
    test_source = tmp_path / "test_source.txt"
    test_link = tmp_path / "test_link.txt"
    test_source.touch()
    test_link.symlink_to(test_source)
    results = win_file.is_link(str(test_link))
    expected = True
    assert results == expected


def test_is_link_false(tmp_path):
    test_file = tmp_path / "test_not_link.txt"
    test_file.touch()
    results = win_file.is_link(str(test_file))
    expected = False
    assert results == expected


def test_mkdir(tmp_path):
    test_dir = tmp_path / "test_dir"
    grant_perms = {"Guests": {"perms": "full_control"}}
    win_file.mkdir(
        path=str(test_dir),
        owner="Administrators",
        grant_perms=grant_perms,
    )
    owner = win_file.get_user(str(test_dir))
    assert owner == "Administrators"
    perms = salt.utils.win_dacl.get_permissions(str(test_dir))
    assert perms["Not Inherited"]["Guests"]["grant"]["permissions"] == "Full control"


def test_check_perms(tmp_path):
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()
    grant_perms = {"Guests": {"perms": "full_control"}}
    ret = {}
    with patch.dict(salt.utils.win_dacl.__opts__, {"test": False}):
        result = win_file.check_perms(
            path=str(test_dir),
            ret=ret,
            owner="Guests",
            grant_perms=grant_perms,
            inheritance=False,
        )

    expected = {
        "changes": {
            "grant_perms": {
                "Guests": {
                    "permissions": "full_control",
                },
            },
            "owner": "Guests",
        },
        "comment": "",
        "name": str(test_dir),
        "result": True,
    }

    assert result["changes"]["grant_perms"] == expected["changes"]["grant_perms"]
    owner = win_file.get_user(str(test_dir))
    assert owner == "Guests"
    perms = salt.utils.win_dacl.get_permissions(str(test_dir))
    assert perms["Not Inherited"]["Guests"]["grant"]["permissions"] == "Full control"


def test_set_perms(tmp_path):
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()
    grant_perms = {"Guests": {"perms": "full_control"}}
    win_file.set_perms(
        path=str(test_dir),
        grant_perms=grant_perms,
    )
    perms = salt.utils.win_dacl.get_permissions(str(test_dir))
    assert perms["Not Inherited"]["Guests"]["grant"]["permissions"] == "Full control"


def test_get_user():
    """
    We can't know what this value is, so we're just making sure it found
    something
    """
    result = win_file.get_user(os.getenv("WINDIR"))
    assert result != ""


def test_get_user_error():
    with pytest.raises(CommandExecutionError):
        win_file.get_user("C:\\fake\\path")


def test_version_missing_file():
    with pytest.raises(CommandExecutionError):
        win_file.version("C:\\Windows\\bogus.exe")


def test_version_missing_directory():
    with pytest.raises(CommandExecutionError):
        win_file.version("C:\\Windows\\System32")


def test_version_details_missing_file():
    with pytest.raises(CommandExecutionError):
        win_file.version_details("C:\\Windows\\bogus.exe")


def test_version_details_missing_directory():
    with pytest.raises(CommandExecutionError):
        win_file.version_details("C:\\Windows\\System32")


def test_version_details_exe():
    result = win_file.version_details(r"C:\Windows\notepad.exe")
    assert result["FileDescription"] == "Notepad"
    assert result["FileType"] == "Application"
    assert result["OperatingSystem"] == ["32-bit Windows", "Windows NT"]
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result["ProductVersion"])
    assert regex.search(result["Version"])


def test_version_details_dll():
    result = win_file.version_details(r"C:\Windows\System32\FirewallAPI.dll")
    assert "Firewall API" in result["FileDescription"]
    assert result["FileType"] == "DLL"
    assert result["OperatingSystem"] == ["32-bit Windows", "Windows NT"]
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result["ProductVersion"])
    assert regex.search(result["Version"])


def test_version_details_sys():
    result = win_file.version_details(r"C:\Windows\System32\drivers\netio.sys")
    assert result["FileDescription"] == "Network I/O Subsystem"
    assert result["FileType"] == "Network Driver"
    assert result["OperatingSystem"] == ["32-bit Windows", "Windows NT"]
    regex = re.compile(r"\d+.\d+.\d+.\d+")
    assert regex.search(result["ProductVersion"])
    assert regex.search(result["Version"])
